import os
import time
import asyncio
import contextlib

from lazyops.imports._sqlalchemy import require_sql

require_sql(required=True, require_asyncpg=True, require_psycopg2=True)

from sqlalchemy import create_engine, event, exc
from sqlalchemy.engine.base import Engine
from sqlalchemy.orm import declarative_base, sessionmaker
from sqlalchemy.orm import Session, scoped_session
from sqlalchemy.pool import NullPool
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine, AsyncEngine, async_scoped_session

from lazyops.utils.logs import logger
from lazyops.utils import Json
from lazyops.types import BaseModel, lazyproperty, BaseSettings, Field
from typing import Any, Generator, AsyncGenerator, Optional, Union, Type, Dict, cast, TYPE_CHECKING, List, Tuple, TypeVar, Callable
from pydantic.networks import PostgresDsn
from lazyops.libs.psqldb.retry import reconnecting_engine
from lazyops.utils.helpers import import_string

Base = declarative_base()

EngineT = Union[Engine, AsyncEngine]
SessionT = Union[Session, AsyncSession]
SettingsT = Union[Type[BaseSettings], Any]

class SafePostgresDsn(PostgresDsn):

    @property
    def safestr(self) -> str:
        """
        Returns the safe string
        """
        s = str(self)
        if self.password: s = s.replace(self.password, '********', 1)
        return s


class Dummy(BaseModel):
    dsn: SafePostgresDsn

def uri_builder(uri: Union[str, SafePostgresDsn], scheme: Optional[str] = None) -> SafePostgresDsn:
    """
    Helper to construct a PostgresDsn from a string
    """
    # print(uri)
    if uri is None: raise ValueError('uri cannot be empty')
    if isinstance(uri, str):
        # if scheme and not uri.startswith(scheme):
        if scheme and '://' not in uri:
            uri = f'{scheme}://{uri}'
        x = Dummy(dsn=uri)
        uri = x.dsn
    
    if scheme and uri.scheme != scheme:
        new = uri.replace(uri.scheme, scheme)
        x = Dummy(dsn=new)
        uri = x.dsn
    
    return uri

def get_nested_arg(
    config: Dict[str, Any],
    base_config: Dict[str, Any],
    key: str,
    default: Any = None,
) -> Any:
    """
    Helper to get a nested value from a config
    """
    return config.pop(key) if key in config else base_config.get(key, default)




class Context(BaseModel):
    uri: SafePostgresDsn

    scheme: Optional[str] = 'postgresql+psycopg2'
    async_scheme: Optional[str] = 'postgresql+asyncpg'
    
    eng: Optional[EngineT] = None
    eng_ro: Optional[EngineT] = None

    sess: Optional[SessionT] = None
    sess_ro: Optional[SessionT] = None

    asess: Optional[SessionT] = None
    asess_ro: Optional[SessionT] = None

    aeng: Optional[EngineT] = None
    aeng_ro: Optional[EngineT] = None

    settings: Optional[SettingsT] = None
    config: Optional[Dict[str, Any]] = Field(default = {})

    ctx: Optional[Dict[str, Any]] = Field(default = {})

    @property
    def debug_enabled(self) -> bool:
        """
        Returns the debug_enabled flag from the settings
        """
        if self.settings:
            return getattr(self.settings, 'debug_enabled', False)
        return os.getenv('DEBUG_ENABLED', 'false') in ['true', 'True', '1']
        # return cast(bool, os.getenv('DEBUG_ENABLED', 'false'))
    
    @property
    def is_verbose(self) -> bool:
        """
        Returns the verbose flag from the settings
        """
        if self.settings:
            if hasattr(self.settings, 'logging'):
                return getattr(self.settings.logging, 'db_verbose', False)
            if hasattr(self.settings, 'db_verbose'):
                return getattr(self.settings, 'db_verbose', False)
            if hasattr(self.settings, 'postgres_verbose'):
                return getattr(self.settings, 'postgres_verbose', False)
        return os.getenv('DB_VERBOSE', 'false') in ['true', 'True', '1']
        # return cast(bool, os.getenv('DB_VERBOSE', 'false'))


    @property
    def is_debug_mode(self) -> bool:
        """
        Returns the custom debug mode flag from the settings
        """
        if 'is_debug_mode' in self.ctx:
            return self.ctx['is_debug_mode']
        if self.settings and hasattr(self.settings, 'postgres_debug_mode'):
            self.ctx['is_debug_mode'] = getattr(self.settings, 'postgres_debug_mode', False)
            return self.ctx['is_debug_mode']
        if self.config.get('is_debug_mode') is not None:
            self.ctx['is_debug_mode'] = self.config.get('is_debug_mode')
            return self.ctx['is_debug_mode']
        return False


    @property
    def retries(self) -> int:
        """
        Returns the retries from the settings
        """
        if self.settings:
            return getattr(self.settings, 'postgres_retry_limit', 5)
        return int(os.getenv('DB_RETRIES', '5'))
    
    @property
    def retry_interval(self) -> float:
        """
        Returns the retry_interval from the settings
        """
        if self.settings:
            return getattr(self.settings, 'postgres_retry_interval', 5.0)
        return float(os.getenv('DB_RETRY_INTERVAL', '5.0'))
        # return cast(float, os.getenv('DB_RETRY_INTERVAL', '5.0'))

    @lazyproperty
    def has_pooler(self) -> bool:
        """
        Returns whether the uri has a pooler
        """
        return 'pooler' in self.uri
    
    @lazyproperty
    def uri_ro(self) -> PostgresDsn:
        """
        Returns the read-only uri
        """
        return uri_builder(self.uri.replace('pooler-rw', 'pooler-ro'))

    @lazyproperty
    def async_uri(self) -> PostgresDsn:
        """
        Returns the async uri
        """
        return uri_builder(self.uri, scheme=self.async_scheme)
    
    @lazyproperty
    def async_uri_ro(self) -> PostgresDsn:
        """
        Returns the async read-only uri
        """
        return uri_builder(self.uri_ro, scheme=self.async_scheme)

    @lazyproperty
    def has_ro(self) -> bool:
        """
        Returns whether there is a RO connection available
        """
        return self.uri != self.uri_ro

    def reset(self):
        """
        Resets the connections
        """
        self.eng = None
        self.eng_ro = None

        self.aeng = None
        self.aeng_ro = None
        
        self.sess = None
        self.sess_ro = None

        self.asess = None
        self.asess_ro = None

    @property
    def engine(self) -> Engine:
        """
        Returns the engine
        """
        if self.eng is None:
            eng_args = self.config.get('engine_args', {})
            self.eng = create_engine(
                self.uri,
                echo = get_nested_arg(eng_args, self.config, 'echo', self.is_verbose),
                future = get_nested_arg(eng_args, self.config, 'future', True),
                pool_pre_ping = get_nested_arg(eng_args, self.config, 'pool_pre_ping', True),
                **eng_args,
        )
        return self.eng

    @property
    def engine_ro(self) -> Engine:
        """
        Returns the read-only engine
        """
        if self.eng_ro is None and self.has_ro:
            eng_args = self.config.get('engine_ro_args', {})
            self.eng_ro = create_engine(
                self.uri_ro,
                echo = get_nested_arg(eng_args, self.config, 'echo', self.is_verbose),
                future = get_nested_arg(eng_args, self.config, 'future', True),
                pool_pre_ping = get_nested_arg(eng_args, self.config, 'pool_pre_ping', True),
                **eng_args,
        )
        return self.eng_ro
    
    @property
    def async_engine(self) -> AsyncEngine:
        """
        Returns the async engine
        """
        
        if self.aeng is None:
            eng_args = self.config.get('async_engine_args', {})
            if poolclass := get_nested_arg(eng_args, self.config, 'poolclass', None):
                if isinstance(poolclass, str):
                    poolclass = import_string(poolclass)
                eng_args['poolclass'] = poolclass
            
            # Handle NullPool
            if 'poolclass' in eng_args and eng_args['poolclass'] is not None and eng_args['poolclass'].__name__ == 'NullPool':
                _ = get_nested_arg(eng_args, self.config, 'pool_size')
            elif pool_size := get_nested_arg(eng_args, self.config, 'pool_size', 50) is not None:
                eng_args['pool_size'] = pool_size

            self.aeng = create_async_engine(
                self.async_uri,
                echo = get_nested_arg(eng_args, self.config, 'echo', self.is_verbose),
                json_serializer = get_nested_arg(eng_args, self.config, 'json_serializer', Json.dumps),
                **eng_args,
        )
        return self.aeng

    @property
    def async_engine_ro(self) -> AsyncEngine:
        """
        Returns the read-only async engine
        """
        if self.aeng_ro is None and self.has_ro:
            eng_args = self.config.get('async_engine_ro_args', {})
            if poolclass := get_nested_arg(eng_args, self.config, 'poolclass', None):
                if isinstance(poolclass, str):
                    poolclass = import_string(poolclass)
                eng_args['poolclass'] = poolclass
            
            # Handle NullPool
            if 'poolclass' in eng_args and eng_args['poolclass'] is not None and eng_args['poolclass'].__name__ == 'NullPool':
                _ = get_nested_arg(eng_args, self.config, 'pool_size')
            elif pool_size := get_nested_arg(eng_args, self.config, 'pool_size', 50) is not None:
                eng_args['pool_size'] = pool_size

            self.aeng_ro = create_async_engine(
                self.async_uri_ro,
                echo = get_nested_arg(eng_args, self.config, 'echo', self.is_verbose),
                json_serializer = get_nested_arg(eng_args, self.config, 'json_serializer', Json.dumps),
                execution_options = get_nested_arg(eng_args, self.config, 'execution_options', 
                    {
                        "isolation_level": "READ COMMITTED",
                        "postgresql_readonly": True,
                        "postgresql_deferrable": True,
                    }
                ),
                **eng_args,
        )
        return self.aeng_ro
    
    @property
    def session(self) -> Session:
        """
        Returns the session
        """
        if self.sess is None:
            sess_args = self.config.get('session_args', {})
            if class_ := get_nested_arg(sess_args, self.config, 'class_', Session):
                if isinstance(class_, str):
                    class_ = import_string(class_)
                sess_args['class_'] = class_
            
            if event_hooks := get_nested_arg(sess_args, self.config, 'event_hooks', []):
                self.ctx['session_hooks'] = event_hooks
            
            if is_scoped_session := get_nested_arg(sess_args, self.config, 'is_scoped_session', False):
                self.ctx['is_scoped_session'] = is_scoped_session
            
            self.sess = sessionmaker(
                bind = self.engine,
                autoflush = get_nested_arg(sess_args, self.config, 'autoflush', True),
                expire_on_commit = get_nested_arg(sess_args, self.config, 'expire_on_commit', False),
                **sess_args,
            )
            if is_scoped_session := self.ctx.get('is_scoped_session'):
                self.sess = scoped_session(self.sess)
            
            if event_hooks := self.ctx.get('session_hooks'):
                for sess_hook in event_hooks:
                    event_name, hook = sess_hook
                    event.listen(self.sess, event_name, hook)
        return self.sess
    
    @property
    def session_ro(self) -> Optional[Session]:
        """
        Returns the read-only session
        """
        if self.sess_ro is None and self.has_ro:
            sess_args = self.config.get('session_ro_args', {})
            if class_ := get_nested_arg(sess_args, self.config, 'class_', Session):
                if isinstance(class_, str):
                    class_ = import_string(class_)
                sess_args['class_'] = class_
            
            if event_hooks := get_nested_arg(sess_args, self.config, 'event_hooks', []):
                self.ctx['session_ro_hooks'] = event_hooks
            

            if is_scoped_session := get_nested_arg(sess_args, self.config, 'is_scoped_session', False):
                self.ctx['is_scoped_session_ro'] = is_scoped_session

            self.sess_ro = sessionmaker(
                bind = self.engine_ro,
                autoflush = get_nested_arg(sess_args, self.config, 'autoflush', True),
                expire_on_commit = get_nested_arg(sess_args, self.config, 'expire_on_commit', False),
                **sess_args,
            )

            if is_scoped_session := self.ctx.get('is_scoped_session_ro'):
                self.sess_ro = scoped_session(self.sess_ro)

            if event_hooks := self.ctx.get('session_ro_hooks'):
                for sess_hook in event_hooks:
                    event_name, hook = sess_hook
                    event.listen(self.sess_ro, event_name, hook)
        return self.sess_ro if self.sess_ro is not None else None
    

    @property
    def async_session(self) -> AsyncSession:
        """
        Returns the async session
        """
        if self.asess is None:
            sess_args = self.config.get('async_session_args', {})
            if class_ := get_nested_arg(sess_args, self.config, 'class_', AsyncSession):
                if isinstance(class_, str):
                    class_ = import_string(class_)
                sess_args['class_'] = class_
            
            if event_hooks := get_nested_arg(sess_args, self.config, 'event_hooks', None):
                self.ctx['async_session_hooks'] = event_hooks
            

            if is_scoped_session := get_nested_arg(sess_args, self.config, 'is_scoped_session', False):
                self.ctx['is_scoped_async_session'] = is_scoped_session

            self.asess: AsyncSession = sessionmaker(
                bind = self.async_engine,
                autoflush = get_nested_arg(sess_args, self.config, 'autoflush', True),
                expire_on_commit = get_nested_arg(sess_args, self.config, 'expire_on_commit', False),
                **sess_args,
            )
            if is_scoped_session := self.ctx.get('is_scoped_async_session'):
                self.asess = async_scoped_session(self.asess, scopefunc = asyncio.current_task)
            
        return self.asess
    
    @property
    def async_session_ro(self) -> Optional[AsyncSession]:
        """
        Returns the read-only async session
        """
        if self.asess_ro is None and self.has_ro:
            sess_args = self.config.get('async_session_ro_args', {})
            if class_ := get_nested_arg(sess_args, self.config, 'class_', AsyncSession):
                if isinstance(class_, str):
                    class_ = import_string(class_)
                sess_args['class_'] = class_
            
            if event_hooks := get_nested_arg(sess_args, self.config, 'event_hooks', []):
                self.ctx['async_session_ro_hooks'] = event_hooks
            

            if is_scoped_session := get_nested_arg(sess_args, self.config, 'is_scoped_session', False):
                self.ctx['is_scoped_async_session_ro'] = is_scoped_session
            
            self.asess_ro = sessionmaker(
                bind = self.async_engine_ro,
                autoflush = get_nested_arg(sess_args, self.config, 'autoflush', True),
                expire_on_commit = get_nested_arg(sess_args, self.config, 'expire_on_commit', False),
                **sess_args,
            )

            if is_scoped_session := self.ctx.get('is_scoped_async_session_ro'):
                self.asess_ro = async_scoped_session(self.asess_ro, scopefunc = asyncio.current_task)

        return self.asess_ro if self.asess_ro is not None else None

    def get_sess(self, ro: Optional[bool] = False, _session: Optional[Session] = None, **kwargs) -> Session:
        """
        Returns a session
        """
        if _session is not None and isinstance(_session, Session) and _session.bind is not None: return _session
        if ro and self.has_ro: return self.session_ro(**kwargs)
        
        # Bug testing for sync sessions
        if self.is_debug_mode:
            if 'sync_sessions' not in self.ctx: self.ctx['sync_sessions'] = 0
            self.ctx['sync_sessions'] += 1
            logger.warning(f"[{self.ctx['sync_sessions']}] Using new sync session")
        return self.session(**kwargs)
    
    def get_async_sess(self, ro: Optional[bool] = False, _session: Optional[AsyncSession] = None, **kwargs) -> AsyncSession:
        """
        Returns an async session
        """
        # if _session is not None and isinstance(_session, AsyncSession) and _session.bind is not None: return _session
        if _session is not None and isinstance(_session, AsyncSession): return _session
        if ro and self.has_ro: 
            _session = self.async_session_ro(**kwargs)
            if event_hooks := self.ctx.get('async_session_ro_hooks'):
                for sess_hook in event_hooks:
                    event_name, hook = sess_hook
                    event.listen(_session.sync_session, event_name, hook)
            return _session

        # Bug testing for async sessions
        if self.is_debug_mode:
            if 'async_sessions' not in self.ctx: self.ctx['async_sessions'] = 0
            self.ctx['async_sessions'] += 1
            logger.warning(f"[{self.ctx['async_sessions']}] Using new async session")
        
        _session = self.async_session(**kwargs)
        if self.ctx.get('async_session_hooks'):
            for sess_hook in self.ctx['async_session_hooks']:
                event_name, hook = sess_hook
                event.listen(_session.sync_session, event_name, hook)
        return _session

    
    
    @classmethod
    def from_uri(
        cls,
        uri: str,
        scheme: Optional[str] = 'postgresql+psycopg2',
        async_scheme: Optional[str] = 'postgresql+asyncpg',
        settings: Optional[SettingsT] = None,
        config: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> 'Context':
        """
        Configures the connection
        """
        uri = uri_builder(uri, scheme=scheme)
        if config is None: config = {}
        if kwargs: config.update(kwargs)
        return cls(uri = uri, async_scheme=async_scheme, settings=settings, config=config)


    @contextlib.contextmanager
    def get_session(
        self, 
        ro: Optional[bool] = False, 
        future: bool = True, 
        session: Optional[Session] = None,
        **kwargs
    ) -> Generator[Session, None, None]:
        """
        Context manager for database session
        """
        # sess: Session = self.get_sess(ro = ro, future = future, **kwargs) if session is None else session
        sess: Session = self.get_sess(ro = ro, future = future, _session = session, **kwargs)
        try:
            yield sess
        finally:
            # Close if the session was created here
            if sess is not None and session is None:
                if self.debug_enabled:
                    logger.debug('Closing Session')
                sess.close()
    
    @contextlib.asynccontextmanager
    async def get_async_session(
        self, 
        ro: Optional[bool] = False, 
        retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        session: Optional[AsyncSession] = None,
        **kwargs
    ) -> Generator[AsyncSession, None, None]:
        """
        Async context manager for database session
        """
        # sess: AsyncSession = None
        
        retries = retries or self.retries
        retry_interval = retry_interval or self.retry_interval
        # sess = self.get_async_sess(ro = ro, **kwargs) if session is None else session
        sess = self.get_async_sess(ro = ro, _session = session, **kwargs)
        raw_dbapi_err: Exception = None
        try:
            for retry in range(retries):
                try:
                    async with sess:
                        yield sess
                    break
                except (exc.DisconnectionError, exc.OperationalError, exc.InterfaceError, exc.PendingRollbackError) as raw_dbapi_err:
                    await sess.rollback()
                    logger.error(f"Retrying Execute after error: {raw_dbapi_err.__class__.__name__}: {raw_dbapi_err}")
                    logger.warning(f"Attempt {retry}/{retries}")
                    if raw_dbapi_err.connection_invalidated:
                        logger.warning("Connection invalidated, resetting")
                        await sess.close()

                    await asyncio.sleep(retry_interval)

        finally:
            # Close if the session was created here
            if sess is not None and session is None:
                if self.debug_enabled:
                    logger.debug('Closing Session')
                await sess.close()
                # await sess.dispose()
    

    def create_all(self, base: Optional[Any] = None):
        """
        Creates all tables
        """
        base = base or Base
        with self.engine.begin() as conn:
            conn.run_sync(base.metadata.create_all)
    
    def drop_all(self, base: Optional[Any] = None):
        """
        Drops all tables
        """
        base = base or Base
        with self.engine.begin() as conn:
            conn.run_sync(base.metadata.drop_all)
    
    async def async_create_all(self, base: Optional[Any] = None):
        """
        Creates all tables
        """
        base = base or Base
        async with self.async_engine.begin() as conn:
            await conn.run_sync(base.metadata.create_all)
    
    async def async_drop_all(self, base: Optional[Any] = None):
        """
        Drops all tables
        """
        base = base or Base
        async with self.async_engine.begin() as conn:
            await conn.run_sync(base.metadata.drop_all)

    
class PostgresDBMeta(type):

    _settings: Optional[SettingsT] = None
    _ctx: Optional[Context] = None
    _uri: Optional[str] = None
    _config: Optional[Dict[str, Any]] = None

    scheme: Optional[str] = 'postgresql+psycopg2'
    async_scheme: Optional[str] = 'postgresql+asyncpg'

    get_settings_callable: Optional[Callable] = None

    @property
    def settings(cls) -> SettingsT:
        """
        The settings for the database
        """
        if cls._settings is None:
            cls._settings = cls.get_settings()
            if cls._settings:
                if getattr(cls._settings, 'postgres_scheme', None):
                    cls.scheme = cls._settings.postgres_scheme
                if getattr(cls._settings, 'postgres_async_scheme', None):
                    cls.async_scheme = cls._settings.postgres_async_scheme
        return cls._settings
    
    def get_settings(cls, settings: Optional[SettingsT] = None) -> SettingsT:
        """
        Helper method to override and get/set the settings
        """
        if settings is not None:
            cls.set_settings(settings)
        elif cls.get_settings_callable is not None:
            cls.set_settings(cls.get_settings_callable())
        return cls._settings

    def set_settings(cls, settings: SettingsT):
        """
        Sets the settings
        """
        cls._settings = settings


    @property
    def uri(cls) -> Union[str, SafePostgresDsn]:
        """
        Returns the uri
        """
        if not cls._uri and cls.settings:
            if getattr(cls.settings, 'pg_uri', None):
                # This should be a PostgresDsn
                cls._uri = cls.settings.pg_uri
                # logger.dev(f"Using pg_uri: {cls._uri}")
            elif getattr(cls.settings, 'postgres_url', None):
                # This should be a string
                # logger.dev(f"Using postgres_url: {cls.settings.postgres_url}")
                cls._uri = uri_builder(cls.settings.postgres_url, scheme=cls.scheme)
            elif postgres_host := getattr(cls.settings, 'postgres_host', os.getenv('POSTGRES_HOST')):
                base_uri = f"{postgres_host}:{getattr(cls.settings, 'postgres_port', 5432)}/{getattr(cls.settings, 'postgres_db', 'postgres')}"
                if postgres_user := getattr(cls.settings, 'postgres_user', os.getenv('POSTGRES_USER')):
                    base_uri = (
                        f"{postgres_user}:{cls.settings.postgres_password}@{base_uri}"
                        if getattr(cls.settings, 'postgres_password', None)
                        else f"{postgres_user}@{base_uri}"
                    )
                # logger.dev(f"Using postgres_host: {base_uri}")
                cls._uri = uri_builder(base_uri, scheme=cls.scheme)
        if not cls._uri:
            cls._uri = uri_builder(os.getenv('POSTGRES_URI', 'postgres@127.0.0.1:5432/postgres'), scheme=cls.scheme)
        return cls._uri
    
    @property
    def safe_uri(cls) -> str:
        """
        Returns the safe uri for logging
        """
        return cls.uri.safestr

    @property
    def pg_admin_user(cls) -> str:
        """
        Returns the admin user
        """
        if admin_user := cls.config.get(
            'pg_admin_user', os.getenv('POSTGRES_ADMIN_USER', os.getenv('POSTGRES_USER')),
        ):
            return admin_user
        return cls.uri.user
    
    @property
    def pg_admin_password(cls) -> str:
        """
        Returns the admin password
        """
        if admin_password := cls.config.get(
            'pg_admin_password', os.getenv('POSTGRES_ADMIN_PASSWORD', os.getenv('POSTGRES_PASSWORD')),
        ):
            return admin_password
        return cls.uri.password
    
    @property
    def pg_admin_db(cls) -> str:
        """
        Returns the admin db
        """
        if admin_db := cls.config.get(
            'pg_admin_db', os.getenv('POSTGRES_ADMIN_DB', os.getenv('POSTGRES_DB')),
        ):
            return admin_db
        return cls.uri.path[1:]
    
    @property
    def admin_uri(cls) -> SafePostgresDsn:
        """
        Returns the admin uri
        """
        uri = f'{cls.uri.host}:{cls.uri.port}/{cls.pg_admin_db}'
        auth = f'{cls.pg_admin_user}'
        if cls.pg_admin_password:
            auth += f':{cls.pg_admin_password}'
        uri = f'{auth}@{uri}'
        return uri_builder(uri, scheme = cls.scheme)
    
    def get_admin_uri(
        cls, 
        host: Optional[str] = None,
        port: Optional[int] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        db: Optional[str] = None,
    ) -> SafePostgresDsn:
        """
        Returns the admin uri
        """
        uri = f'{host or cls.uri.host}:{port or cls.uri.port}/{db or cls.pg_admin_db}'
        auth = f'{user or cls.pg_admin_user}'
        if password or cls.pg_admin_password:
            auth += f':{password or cls.pg_admin_password}'
        uri = f'{auth}@{uri}'
        return uri_builder(uri, scheme = cls.scheme)

    
    @property
    def config(cls) -> Dict[str, Any]:
        """
        Returns the config
        """
        if not cls._config:
            cls._config = {}
            if cls.settings:
                if getattr(cls.settings, 'pg_config', None):
                    cls._config = cls.settings.pg_config
                elif getattr(cls.settings, 'postgres_config', None):
                    cls._config = cls.settings.postgres_config
        return cls._config
    
    @property
    def ctx(cls) -> Context:
        """
        Returns the context
        """
        if not cls._ctx:
            cls._ctx = Context.from_uri(
                cls.uri, 
                settings = cls.settings, 
                config = cls.config, 
                scheme = cls.scheme, 
                async_scheme = cls.async_scheme
            )
        return cls._ctx

    @property
    def engine(cls) -> Engine:
        """
        Returns the read-write engine
        """
        return cls.ctx.engine
    
    @property
    def engine_ro(cls) -> Optional[Engine]:
        """
        Returns the read-only engine
        """
        return cls.ctx.engine_ro
    
    @property
    def async_engine(cls) -> AsyncEngine:
        """
        Returns the read-write async engine
        """
        return cls.ctx.async_engine
    
    @property
    def async_engine_ro(cls) -> Optional[AsyncEngine]:
        """
        Returns the read-only async engine
        """
        return cls.ctx.async_engine_ro
    
    # @property
    # def session(cls) -> Session:
    #     """
    #     Returns the read-write session
    #     """
    #     return cls.ctx.session
    
    # @property
    # def session_ro(cls) -> Optional[Session]:
    #     """
    #     Returns the read-only session
    #     """
    #     return cls.ctx.session_ro
    
    # @property
    # def async_session(cls) -> AsyncSession:
    #     """
    #     Returns the read-write async session
    #     """
    #     return cls.ctx.async_session
    
    # @property
    # def async_session_ro(cls) -> Optional[AsyncSession]:
    #     """
    #     Returns the read-only async session
    #     """
    #     return cls.ctx.async_session_ro
    
    def get_sess(cls, ro: Optional[bool] = False, future: bool = True, **kwargs) -> Session:
        """
        Returns a session
        """
        return cls.ctx.get_sess(ro=ro, future=future, **kwargs)
    
    def get_async_sess(cls, ro: Optional[bool] = False, **kwargs) -> AsyncSession:
        """
        Returns an async session
        """
        return cls.ctx.get_async_sess(ro=ro, **kwargs)
    

    @contextlib.contextmanager
    def session(
        cls, 
        ro: Optional[bool] = False, 
        future: bool = True, 
        session: Optional[Session] = None,
        **kwargs
    ) -> Generator[Session, None, None]:
        """
        Context manager for database session
        """
        # if session is not None:
        #     yield session
        # else:
        with cls.ctx.get_session(ro = ro, future = future, session = session, **kwargs) as sess:
            yield sess
    
    @contextlib.asynccontextmanager
    async def async_session(
        cls, 
        ro: Optional[bool] = False, 
        retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        session: Optional[AsyncSession] = None,
        **kwargs
    ) -> Generator[AsyncSession, None, None]:
        """
        Async context manager for database session
        """
        # if session is not None:
        #     yield session
        # else:
        async with cls.ctx.get_async_session(ro = ro, retries = retries, retry_interval = retry_interval, session = session, **kwargs) as sess:
            yield sess

    
    def create_all(cls, base: Optional[Any] = None):
        """
        Creates all tables
        """
        base = base or Base
        return cls.ctx.create_all(base=base)
    
    def drop_all(cls, base: Optional[Any] = None):
        """
        Drops all tables
        """
        base = base or Base
        return cls.ctx.drop_all(base=base)
    
    async def async_create_all(cls, base: Optional[Any] = None):
        """
        Creates all tables
        """
        base = base or Base
        return await cls.ctx.async_create_all(base=base)
    
    async def async_drop_all(cls, base: Optional[Any] = None):
        """
        Drops all tables
        """
        base = base or Base
        return await cls.ctx.async_drop_all(base=base)
    

if TYPE_CHECKING:
    from lazyops.libs.dbinit import Privilege

    
class PostgresDB(metaclass=PostgresDBMeta):
    """
    Postgres database
    """
    
    @classmethod
    def prepare_db(
        cls, 

        role_user: Optional[str] = None,
        role_password: Optional[str] = None,
        role_createdb: Optional[bool] = True,
        role_createrole: Optional[bool] = True,
        role_replication: Optional[bool] = True,
        role_bypassrls: Optional[bool] = True,
        role_login: Optional[bool] = None,
        role_options: Optional[Dict[str, Any]] = None,
        role_superuser: Optional[bool] = True,


        db_name: Optional[str] = None,
        db_owner: Optional[str] = None,
        db_grants: Optional[List[Tuple[str, Optional['Privilege']]]] = None,
        db_options: Optional[Dict[str, Any]] = None,

        db_admin_uri: Optional[Union[str, PostgresDsn]] = None,

        statements: Optional[List[str]] = None,
        grant_public_schema: Optional[bool] = True,
        verbose: Optional[bool] = True,
        raise_exception: Optional[bool] = False,
        **kwargs,

    ):  # sourcery skip: low-code-quality
        """
        Runs initialization for the DB
        """
        from lazyops.libs.dbinit import (
            Controller,
            GrantTo,
            Privilege,
            Database,
            Role,
        )

        pg_user = role_user or cls.ctx.uri.user
        pg_pass = role_password or cls.ctx.uri.password
        if role_login is None: role_login = bool(pg_pass)
        if not role_options: role_options = {}
        if not db_options: db_options = {}

        if pg_user == "postgres":
            role = Role(name = pg_user)
        else:
            role = Role(
                name = pg_user,
                createdb = role_createdb,
                createrole = role_createrole,
                replication = role_replication,
                bypassrls = role_bypassrls,
                login = role_login,
                password = pg_pass,
                superuser = role_superuser,
                **role_options
            )
        if verbose: logger.info(f"Creating Role {pg_user}: {role}")
        pg_db = db_name or cls.ctx.uri.path[1:]
        db_owner = role if db_owner is None else Role(name = db_owner)
        if not db_grants: 
            db_grants = [
                GrantTo(
                    privileges = [Privilege.ALL_PRIVILEGES], to = [role]
                )
            ]

        db = Database(
            name = pg_db,
            owner = db_owner,
            grants = db_grants,
            **db_options
        )
        if verbose: logger.info(f"Creating Database {pg_db}: {db}")
        db_admin_uri = db_admin_uri or cls.admin_uri
        try:
            Controller.run_all(engine = create_engine(url = db_admin_uri))
        except Exception as e:
            logger.trace(f"Error creating database: {pg_db}", error = e)
            if raise_exception: raise e

        # engine = create_engine(url = cls.ctx.uri)
        engine = create_engine(url = cls.get_admin_uri(db = pg_db))
        if statements:
            for statement in statements:
                if verbose: logger.info(f"Executing statement: {statement}")
                try:
                    engine.execute(statement = statement)
                except Exception as e:
                    logger.trace(f"Error executing statement: {statement}", error = e)
                    if raise_exception: raise e
        
        if grant_public_schema:
            pg_user_encoded = f'"{pg_user}"' if '_' in pg_user or '-' in pg_user else pg_user
            schema_statements = [
                f"GRANT ALL PRIVILEGES ON SCHEMA public TO {pg_user_encoded}",
                f"ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT ALL PRIVILEGES ON TABLES TO {pg_user_encoded}",
                f"GRANT ALL PRIVILEGES ON ALL TABLES IN SCHEMA public TO {pg_user_encoded}",
                f"GRANT ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA public TO {pg_user_encoded}",
            ]
            if statements: schema_statements = [statement for statement in schema_statements if statement not in statements]
            for statement in schema_statements:
                if verbose: logger.info(f"Executing statement: {statement}")
                try:
                    engine.execute(statement = statement)
                except Exception as e:
                    logger.trace(f"Error executing statement: {statement}", error = e)
                    if raise_exception: raise e

        if verbose: logger.info("Completed DB initialization")



    
    




"""
Import Handler for `magic`
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import magic
    _magic_available = True
except ImportError:
    magic = object
    _magic_available = False

def resolve_magic(
    required: bool = False,
):
    """
    Ensures that `magic` is available
    """
    global magic, _magic_available
    if not _magic_available:
        resolve_missing('magic', 'python-magic', required = required)
        import magic
        _magic_available = True
        globals()['magic'] = magic


def require_magic(
    required: bool = False,
):
    """
    Wrapper for `resolve_magic` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_magic, 
            func = func, 
            required = required
        )
    return decorator


import aiocache
from lazyops.lazyio.models import LazyJson, LazyPickler


from aiocache import caches
from aiocache.serializers import StringSerializer, PickleSerializer, JsonSerializer

caches.set_config({
    'default': {
        'cache': "aiocache.SimpleMemoryCache",
        'serializer': {
            'class': "lazyops.lazyio.LazyJson"
        }
    },
})

from aiocache import cached as async_cache




import lazyops.libs.sqlcache.base
import os
import io
import zlib
import gzip
import codecs
import struct
import sqlite3
import errno
import pickletools
import dill as pkl
import os.path as op
import functools as ft
import contextlib as cl

from fileio.lib.types import File, FileLike
from typing import Any, Iterable, AsyncIterable, Optional
from lazyops.utils.pooler import ThreadPooler
from lazyops.libs.sqlcache.types import BaseMedium
from lazyops.libs.sqlcache.constants import (
    UNKNOWN,
    MODE_NONE,
    MODE_RAW,
    MODE_TEXT,
    MODE_BINARY,
    MODE_PICKLE,
    KeyT,
    ValueT,
    WindowsExceptionError,
)


class Disk(BaseMedium):
    """
    Disk Based Medium 
    
    Store key and value serialization for SQLite database and files.
    """

    def __init__(
        self, 
        directory: Optional[FileLike] = None, 
        min_file_size: int = 0, 
        pickle_protocol: int = 3, 
        **config
    ):
        """Initialize disk instance.
        :param str directory: directory path
        :param int min_file_size: minimum size for file use
        :param int pickle_protocol: pickle protocol for serialization
        :param dict config: dict to pass to update sql settings
        """
        self._directory = File(directory) or File.get_tempdir(prefix = 'lazyops')
        self.min_file_size = min_file_size
        self.pickle_protocol = pickle_protocol
        self.config = config


    def store(self, value: ValueT, read: bool, key: KeyT = UNKNOWN):
        """Convert `value` to fields size, mode, filename, and value for Store
        table.
        :param value: value to convert
        :param bool read: True when value is file-like object
        :param key: key for item (default UNKNOWN)
        :return: (size, mode, filename, value) tuple for Store table
        """
        # pylint: disable=unidiomatic-typecheck
        type_value = type(value)
        min_file_size = self.min_file_size

        if (
            (type_value is str and len(value) < min_file_size)
            or (
                type_value is int
                and -9223372036854775808 <= value <= 9223372036854775807
            )
            or (type_value is float)
        ):
            return 0, MODE_RAW, None, value
        elif type_value is bytes:
            if len(value) < min_file_size:
                return 0, MODE_RAW, None, sqlite3.Binary(value)
            filename, full_path = self.filename(key, value)
            self._write(full_path, io.BytesIO(value), 'xb')
            return len(value), MODE_BINARY, filename, None
        elif type_value is str:
            filename, full_path = self.filename(key, value)
            self._write(full_path, io.StringIO(value), 'x', 'UTF-8')
            # size = op.getsize(full_path)
            size = full_path.size()
            # size = (full_path.info())['size']
            return size, MODE_TEXT, filename, None
        elif read:
            reader = ft.partial(value.read, 2**22)
            filename, full_path = self.filename(key, value)
            iterator = iter(reader, b'')
            size = self._write(full_path, iterator, 'xb')
            return size, MODE_BINARY, filename, None

        result = self.serialize(result)
        if len(result) < min_file_size:
            return 0, MODE_PICKLE, None, sqlite3.Binary(result)
        filename, full_path = self.filename(key, value)
        self._write(full_path, io.BytesIO(result), 'xb')
        return len(result), MODE_PICKLE, filename, None
    
    async def astore(self, value: ValueT, read: bool, key: KeyT = UNKNOWN):
        """Convert `value` to fields size, mode, filename, and value for Store
        table.
        :param value: value to convert
        :param bool read: True when value is file-like object
        :param key: key for item (default UNKNOWN)
        :return: (size, mode, filename, value) tuple for Store table
        """
        # pylint: disable=unidiomatic-typecheck
        type_value = type(value)
        min_file_size = self.min_file_size

        if (
            (type_value is str and len(value) < min_file_size)
            or (
                type_value is int
                and -9223372036854775808 <= value <= 9223372036854775807
            )
            or (type_value is float)
        ):
            return 0, MODE_RAW, None, value
        elif type_value is bytes:
            if len(value) < min_file_size:
                return 0, MODE_RAW, None, sqlite3.Binary(value)
            filename, full_path = self.filename(key, value)
            await self._awrite(full_path, io.BytesIO(value), 'xb')
            return len(value), MODE_BINARY, filename, None
        elif type_value is str:
            filename, full_path = self.filename(key, value)
            await self._awrite(full_path, io.StringIO(value), 'x', 'UTF-8')
            size = await full_path.async_size()
            # size = (await full_path.async_info())['size']
            # size = op.getsize(full_path)
            return size, MODE_TEXT, filename, None
        elif read:
            reader = ft.partial(value.read, 2**22)
            filename, full_path = self.filename(key, value)
            iterator = iter(reader, b'')
            size = await self._awrite(full_path, iterator, 'xb')
            return size, MODE_BINARY, filename, None

        result = await self.aserialize(value)
        # result = pkl.dumps(value, protocol=self.pickle_protocol)
        if len(result) < min_file_size:
            return 0, MODE_PICKLE, None, sqlite3.Binary(result)
        filename, full_path = self.filename(key, value)
        await self._awrite(full_path, io.BytesIO(result), 'xb')
        return len(result), MODE_PICKLE, filename, None
    
    def _write(self, full_path: FileLike, iterator: Iterable, mode: str, encoding: str = None):
        # full_dir, _ = op.split(full_path)
        full_dir = full_path.parent
        for count in range(1, 11):
            with cl.suppress(OSError):
                full_dir.mkdir(parents=True, exist_ok=True)
                # os.makedirs(full_dir)

            try:
                # Another cache may have deleted the directory before
                # the file could be opened.
                writer = full_path.open(mode = mode, encoding = encoding)
                # writer = open(full_path, mode, encoding=encoding)
            except OSError:
                if count == 10:
                    # Give up after 10 tries to open the file.
                    raise
                continue

            with writer:
                size = 0
                for chunk in iterator:
                    size += len(chunk)
                    writer.write(chunk)
                return size
    
    async def _awrite(self, full_path: FileLike, iterator: Iterable, mode: str, encoding: str = None):
        full_dir = full_path.parent
        for count in range(1, 11):
            with cl.suppress(OSError):
                full_dir.mkdir(parents=True, exist_ok=True)

            try:
                # Another cache may have deleted the directory before
                # the file could be opened.
                writer = full_path.async_open(mode = mode, encoding = encoding)
            except OSError:
                if count == 10:
                    # Give up after 10 tries to open the file.
                    raise
                continue

            async with writer:
                size = 0
                for chunk in iterator:
                    size += len(chunk)
                    await writer.write(chunk)
                return size

    def fetch(self, mode: int, filename: str, value: ValueT, read: bool):
        """Convert fields `mode`, `filename`, and `value` from Store table to
        value.
        :param int mode: value mode raw, binary, text, or pickle
        :param str filename: filename of corresponding value
        :param value: database value
        :param bool read: when True, return an open file handle
        :return: corresponding Python value
        """
        # pylint: disable=no-self-use,unidiomatic-typecheck
        if mode == MODE_RAW:
            return bytes(value) if type(value) is sqlite3.Binary else value
        elif mode == MODE_BINARY:
            if read:
                return self._directory.joinpath(filename).open('rb')
                # return open(op.join(self._directory, filename), 'rb')
            return self._directory.joinpath(filename).read_bytes()
            # with open(op.join(self._directory, filename), 'rb') as reader:
            #     return reader.read()
        elif mode == MODE_TEXT:
            return self._directory.joinpath(filename).read_text(encoding='UTF-8')
            # full_path = op.join(self._directory, filename)
            # with open(full_path, 'r', encoding='UTF-8') as reader:
            #     return reader.read()
        elif mode == MODE_PICKLE:
            if value is not None: return self.deserialize(io.BytesIO(value))
            return self.deserialize(self._directory.joinpath(filename).read_bytes())
            # with open(op.join(self._directory, filename), 'rb') as reader:
            #     return pkl.load(reader)

    async def afetch(self, mode: int, filename: str, value: ValueT, read: bool):
        """Convert fields `mode`, `filename`, and `value` from Store table to
        value.
        :param int mode: value mode raw, binary, text, or pickle
        :param str filename: filename of corresponding value
        :param value: database value
        :param bool read: when True, return an open file handle
        :return: corresponding Python value
        """
        if mode == MODE_RAW:
            return bytes(value) if type(value) is sqlite3.Binary else value
        elif mode == MODE_BINARY:
            if read: return self._directory.joinpath(filename).async_open('rb')
            return await self._directory.joinpath(filename).async_read_bytes()
        
        elif mode == MODE_TEXT:
            return await self._directory.joinpath(filename).async_read_text(encoding='UTF-8')

        elif mode == MODE_PICKLE:
            if value is not None: return await self.adeserialize(io.BytesIO(value))
            return await self.adeserialize(await self._directory.joinpath(filename).async_read_bytes())

    def filename(self, key: KeyT = UNKNOWN, value: ValueT = UNKNOWN):
        """Return filename and full-path tuple for file storage.
        Filename will be a randomly generated 28 character hexadecimal string
        with ".val" suffixed. Two levels of sub-directories will be used to
        reduce the size of directories. On older filesystems, lookups in
        directories with many files may be slow.
        The default implementation ignores the `key` and `value` parameters.
        In some scenarios, for example :meth:`Store.push
        <diskcache.Store.push>`, the `key` or `value` may not be known when the
        item is stored in the cache.
        :param key: key for item (default UNKNOWN)
        :param value: value for item (default UNKNOWN)
        """
        # pylint: disable=unused-argument
        hex_name = codecs.encode(os.urandom(16), 'hex').decode('utf-8')
        sub_dir = op.join(hex_name[:2], hex_name[2:4])
        name = f'{hex_name[4:]}.val'
        directory = self._directory.joinpath(sub_dir)
        # directory = op.join(self._directory, sub_dir)
        try:
            directory.mkdir(parents=True, exist_ok=True)
            # os.makedirs(directory)
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise
        
        filename = op.join(sub_dir, name)
        full_path = self._directory.joinpath(filename)
        # full_path = op.join(self._directory, filename)
        return filename, full_path

    def remove(self, filename: str):
        """Remove a file given by `filename`.
        This method is cross-thread and cross-process safe. If an "error no
        entry" occurs, it is suppressed.
        :param str filename: relative path to file
        """
        # full_path = op.join(self._directory, filename)
        full_path = self._directory.joinpath(filename)
        try:
            # os.remove(full_path)
            full_path.unlink()

        except WindowsExceptionError:
            pass
        except OSError as error:
            if error.errno != errno.ENOENT:
                # ENOENT may occur if two caches attempt to delete the same
                # file at the same time.
                raise

    async def aremove(self, filename: str):
        """Remove a file given by `filename`.
        This method is cross-thread and cross-process safe. If an "error no
        entry" occurs, it is suppressed.
        :param str filename: relative path to file
        """
        # full_path = op.join(self._directory, filename)
        full_path = self._directory.joinpath(filename)
        try:
            # os.remove(full_path)
            await full_path.async_unlink()

        except WindowsExceptionError:
            pass
        except OSError as error:
            if error.errno != errno.ENOENT:
                # ENOENT may occur if two caches attempt to delete the same
                # file at the same time.
                raise


class CompressedDisk(Disk):
    "Store key and value using with zlib compression."

    def __init__(
        self, 
        directory: Optional[FileLike] = None, 
        compression_level: int = 3, 
        **config
    ):
        """Initialize Pickle disk instance.
        Keys and values are compressed using the zlib library. The
        `compress_level` is an integer from 0 to 9 controlling the level of
        compression; 1 is fastest and produces the least compression, 9 is
        slowest and produces the most compression, and 0 is no compression.
        :param str directory: directory path
        :param int compression_level: zlib compression level (default 3)
        :param kwargs: super class arguments
        """
        self.compression_level = compression_level
        super().__init__(directory = directory, **config)

    
    def serialize(self, value: ValueT, protocol: Optional[Any] = None, optimize: Optional[bool] = False, **kwargs) -> bytes:
        """
        Serialize value to bytes
        """
        protocol = protocol or self.pickle_protocol
        data = pkl.dumps(value, protocol=self.pickle_protocol)
        if optimize: data = pickletools.optimize(data)
        data = zlib.compress(data, self.compression_level)
        return data
    

    def deserialize(self, data: bytes, **kwargs) -> ValueT:
        """
        Deserialize bytes to value
        """
        return pkl.loads(zlib.decompress(data))




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.text_format import Merge
try:
    from fileio.src import get_pathlike
except:
    from fileio import get_pathlike

from ._base import TFSModelConfig, TFSModelVersion, List,  Union, Optional, Any, Dict
from ._base import dataclass, lazyclass
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='config.proto',
  package='serving_config',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0c\x63onfig.proto\x12\x0eserving_config\x1a\x19google/protobuf/any.proto\"\xa6\x06\n!FileSystemStoragePathSourceConfig\x12V\n\tservables\x18\x05 \x03(\x0b\x32\x43.serving_config.FileSystemStoragePathSourceConfig.ServableToMonitor\x12\x19\n\rservable_name\x18\x01 \x01(\tB\x02\x18\x01\x12\x15\n\tbase_path\x18\x02 \x01(\tB\x02\x18\x01\x12%\n\x1d\x66ile_system_poll_wait_seconds\x18\x03 \x01(\x03\x12,\n fail_if_zero_versions_at_startup\x18\x04 \x01(\x08\x42\x02\x18\x01\x12(\n servable_versions_always_present\x18\x06 \x01(\x08\x1a\x91\x03\n\x15ServableVersionPolicy\x12`\n\x06latest\x18\x64 \x01(\x0b\x32N.serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.LatestH\x00\x12Z\n\x03\x61ll\x18\x65 \x01(\x0b\x32K.serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.AllH\x00\x12\x64\n\x08specific\x18\x66 \x01(\x0b\x32P.serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.SpecificH\x00\x1a\x1e\n\x06Latest\x12\x14\n\x0cnum_versions\x18\x01 \x01(\r\x1a\x05\n\x03\x41ll\x1a\x1c\n\x08Specific\x12\x10\n\x08versions\x18\x01 \x03(\x03\x42\x0f\n\rpolicy_choice\x1a\x64\n\x11ServableToMonitor\x12\x15\n\rservable_name\x18\x01 \x01(\t\x12\x11\n\tbase_path\x18\x02 \x01(\t\x12\x1f\n\x17servable_version_policy\x18\x04 \x01(\tJ\x04\x08\x03\x10\x04\"\xea\x02\n\x0bModelConfig\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tbase_path\x18\x02 \x01(\t\x12\x31\n\nmodel_type\x18\x03 \x01(\x0e\x32\x19.serving_config.ModelTypeB\x02\x18\x01\x12\x16\n\x0emodel_platform\x18\x04 \x01(\t\x12\x65\n\x14model_version_policy\x18\x07 \x01(\x0b\x32G.serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy\x12\x46\n\x0eversion_labels\x18\x08 \x03(\x0b\x32..serving_config.ModelConfig.VersionLabelsEntry\x1a\x34\n\x12VersionLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03:\x02\x38\x01J\x04\x08\x05\x10\x06J\x04\x08\t\x10\n\">\n\x0fModelConfigList\x12+\n\x06\x63onfig\x18\x01 \x03(\x0b\x32\x1b.serving_config.ModelConfig\"\x90\x01\n\x11ModelServerConfig\x12<\n\x11model_config_list\x18\x01 \x01(\x0b\x32\x1f.serving_config.ModelConfigListH\x00\x12\x33\n\x13\x63ustom_model_config\x18\x02 \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x42\x08\n\x06\x63onfig*N\n\tModelType\x12\x1e\n\x16MODEL_TYPE_UNSPECIFIED\x10\x00\x1a\x02\x08\x01\x12\x12\n\nTENSORFLOW\x10\x01\x1a\x02\x08\x01\x12\r\n\x05OTHER\x10\x02\x1a\x02\x08\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])

_MODELTYPE = _descriptor.EnumDescriptor(
  name='ModelType',
  full_name='serving_config.ModelType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MODEL_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=b'\010\001',
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TENSORFLOW', index=1, number=1,
      serialized_options=b'\010\001',
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='OTHER', index=2, number=2,
      serialized_options=b'\010\001',
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1444,
  serialized_end=1522,
)
_sym_db.RegisterEnumDescriptor(_MODELTYPE)

ModelType = enum_type_wrapper.EnumTypeWrapper(_MODELTYPE)
MODEL_TYPE_UNSPECIFIED = 0
TENSORFLOW = 1
OTHER = 2



_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_LATEST = _descriptor.Descriptor(
  name='Latest',
  full_name='serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Latest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_versions', full_name='serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Latest.num_versions', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=680,
  serialized_end=710,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_ALL = _descriptor.Descriptor(
  name='All',
  full_name='serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.All',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=712,
  serialized_end=717,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_SPECIFIC = _descriptor.Descriptor(
  name='Specific',
  full_name='serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Specific',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='versions', full_name='serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Specific.versions', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=719,
  serialized_end=747,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY = _descriptor.Descriptor(
  name='ServableVersionPolicy',
  full_name='serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='latest', full_name='serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.latest', index=0,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all', full_name='serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.all', index=1,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='specific', full_name='serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.specific', index=2,
      number=102, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_LATEST, _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_ALL, _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_SPECIFIC, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='policy_choice', full_name='serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.policy_choice',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=363,
  serialized_end=764,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR = _descriptor.Descriptor(
  name='ServableToMonitor',
  full_name='serving_config.FileSystemStoragePathSourceConfig.ServableToMonitor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='servable_name', full_name='serving_config.FileSystemStoragePathSourceConfig.ServableToMonitor.servable_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_path', full_name='serving_config.FileSystemStoragePathSourceConfig.ServableToMonitor.base_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='servable_version_policy', full_name='serving_config.FileSystemStoragePathSourceConfig.ServableToMonitor.servable_version_policy', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=766,
  serialized_end=866,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG = _descriptor.Descriptor(
  name='FileSystemStoragePathSourceConfig',
  full_name='serving_config.FileSystemStoragePathSourceConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='servables', full_name='serving_config.FileSystemStoragePathSourceConfig.servables', index=0,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='servable_name', full_name='serving_config.FileSystemStoragePathSourceConfig.servable_name', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_path', full_name='serving_config.FileSystemStoragePathSourceConfig.base_path', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file_system_poll_wait_seconds', full_name='serving_config.FileSystemStoragePathSourceConfig.file_system_poll_wait_seconds', index=3,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fail_if_zero_versions_at_startup', full_name='serving_config.FileSystemStoragePathSourceConfig.fail_if_zero_versions_at_startup', index=4,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='servable_versions_always_present', full_name='serving_config.FileSystemStoragePathSourceConfig.servable_versions_always_present', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY, _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=60,
  serialized_end=866,
)


_MODELCONFIG_VERSIONLABELSENTRY = _descriptor.Descriptor(
  name='VersionLabelsEntry',
  full_name='serving_config.ModelConfig.VersionLabelsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='serving_config.ModelConfig.VersionLabelsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='serving_config.ModelConfig.VersionLabelsEntry.value', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1167,
  serialized_end=1219,
)

_MODELCONFIG = _descriptor.Descriptor(
  name='ModelConfig',
  full_name='serving_config.ModelConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='serving_config.ModelConfig.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_path', full_name='serving_config.ModelConfig.base_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_type', full_name='serving_config.ModelConfig.model_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_platform', full_name='serving_config.ModelConfig.model_platform', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version_policy', full_name='serving_config.ModelConfig.model_version_policy', index=4,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version_labels', full_name='serving_config.ModelConfig.version_labels', index=5,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MODELCONFIG_VERSIONLABELSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=869,
  serialized_end=1231,
)


_MODELCONFIGLIST = _descriptor.Descriptor(
  name='ModelConfigList',
  full_name='serving_config.ModelConfigList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='serving_config.ModelConfigList.config', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1233,
  serialized_end=1295,
)


_MODELSERVERCONFIG = _descriptor.Descriptor(
  name='ModelServerConfig',
  full_name='serving_config.ModelServerConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_config_list', full_name='serving_config.ModelServerConfig.model_config_list', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='custom_model_config', full_name='serving_config.ModelServerConfig.custom_model_config', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='config', full_name='serving_config.ModelServerConfig.config',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1298,
  serialized_end=1442,
)

_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_LATEST.containing_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_ALL.containing_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_SPECIFIC.containing_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['latest'].message_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_LATEST
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['all'].message_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_ALL
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['specific'].message_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_SPECIFIC
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.containing_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice'].fields.append(
  _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['latest'])
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['latest'].containing_oneof = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice']
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice'].fields.append(
  _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['all'])
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['all'].containing_oneof = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice']
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice'].fields.append(
  _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['specific'])
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.fields_by_name['specific'].containing_oneof = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY.oneofs_by_name['policy_choice']
_FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR.containing_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG
_FILESYSTEMSTORAGEPATHSOURCECONFIG.fields_by_name['servables'].message_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR
_MODELCONFIG_VERSIONLABELSENTRY.containing_type = _MODELCONFIG
_MODELCONFIG.fields_by_name['model_type'].enum_type = _MODELTYPE
_MODELCONFIG.fields_by_name['model_version_policy'].message_type = _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY
_MODELCONFIG.fields_by_name['version_labels'].message_type = _MODELCONFIG_VERSIONLABELSENTRY
_MODELCONFIGLIST.fields_by_name['config'].message_type = _MODELCONFIG
_MODELSERVERCONFIG.fields_by_name['model_config_list'].message_type = _MODELCONFIGLIST
_MODELSERVERCONFIG.fields_by_name['custom_model_config'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_MODELSERVERCONFIG.oneofs_by_name['config'].fields.append(
  _MODELSERVERCONFIG.fields_by_name['model_config_list'])
_MODELSERVERCONFIG.fields_by_name['model_config_list'].containing_oneof = _MODELSERVERCONFIG.oneofs_by_name['config']
_MODELSERVERCONFIG.oneofs_by_name['config'].fields.append(
  _MODELSERVERCONFIG.fields_by_name['custom_model_config'])
_MODELSERVERCONFIG.fields_by_name['custom_model_config'].containing_oneof = _MODELSERVERCONFIG.oneofs_by_name['config']
DESCRIPTOR.message_types_by_name['FileSystemStoragePathSourceConfig'] = _FILESYSTEMSTORAGEPATHSOURCECONFIG
DESCRIPTOR.message_types_by_name['ModelConfig'] = _MODELCONFIG
DESCRIPTOR.message_types_by_name['ModelConfigList'] = _MODELCONFIGLIST
DESCRIPTOR.message_types_by_name['ModelServerConfig'] = _MODELSERVERCONFIG
DESCRIPTOR.enum_types_by_name['ModelType'] = _MODELTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FileSystemStoragePathSourceConfig = _reflection.GeneratedProtocolMessageType('FileSystemStoragePathSourceConfig', (_message.Message,), {

  'ServableVersionPolicy' : _reflection.GeneratedProtocolMessageType('ServableVersionPolicy', (_message.Message,), {

    'Latest' : _reflection.GeneratedProtocolMessageType('Latest', (_message.Message,), {
      'DESCRIPTOR' : _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_LATEST,
      '__module__' : 'config_pb2'
      # @@protoc_insertion_point(class_scope:serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Latest)
      })
    ,

    'All' : _reflection.GeneratedProtocolMessageType('All', (_message.Message,), {
      'DESCRIPTOR' : _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_ALL,
      '__module__' : 'config_pb2'
      # @@protoc_insertion_point(class_scope:serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.All)
      })
    ,

    'Specific' : _reflection.GeneratedProtocolMessageType('Specific', (_message.Message,), {
      'DESCRIPTOR' : _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY_SPECIFIC,
      '__module__' : 'config_pb2'
      # @@protoc_insertion_point(class_scope:serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy.Specific)
      })
    ,
    'DESCRIPTOR' : _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLEVERSIONPOLICY,
    '__module__' : 'config_pb2'
    # @@protoc_insertion_point(class_scope:serving_config.FileSystemStoragePathSourceConfig.ServableVersionPolicy)
    })
  ,

  'ServableToMonitor' : _reflection.GeneratedProtocolMessageType('ServableToMonitor', (_message.Message,), {
    'DESCRIPTOR' : _FILESYSTEMSTORAGEPATHSOURCECONFIG_SERVABLETOMONITOR,
    '__module__' : 'config_pb2'
    # @@protoc_insertion_point(class_scope:serving_config.FileSystemStoragePathSourceConfig.ServableToMonitor)
    })
  ,
  'DESCRIPTOR' : _FILESYSTEMSTORAGEPATHSOURCECONFIG,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:serving_config.FileSystemStoragePathSourceConfig)
  })
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig)
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig.ServableVersionPolicy)
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig.ServableVersionPolicy.Latest)
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig.ServableVersionPolicy.All)
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig.ServableVersionPolicy.Specific)
_sym_db.RegisterMessage(FileSystemStoragePathSourceConfig.ServableToMonitor)

ModelConfig = _reflection.GeneratedProtocolMessageType('ModelConfig', (_message.Message,), {

  'VersionLabelsEntry' : _reflection.GeneratedProtocolMessageType('VersionLabelsEntry', (_message.Message,), {
    'DESCRIPTOR' : _MODELCONFIG_VERSIONLABELSENTRY,
    '__module__' : 'config_pb2'
    # @@protoc_insertion_point(class_scope:serving_config.ModelConfig.VersionLabelsEntry)
    })
  ,
  'DESCRIPTOR' : _MODELCONFIG,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:serving_config.ModelConfig)
  })
_sym_db.RegisterMessage(ModelConfig)
_sym_db.RegisterMessage(ModelConfig.VersionLabelsEntry)

ModelConfigList = _reflection.GeneratedProtocolMessageType('ModelConfigList', (_message.Message,), {
  'DESCRIPTOR' : _MODELCONFIGLIST,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:serving_config.ModelConfigList)
  })
_sym_db.RegisterMessage(ModelConfigList)

ModelServerConfig = _reflection.GeneratedProtocolMessageType('ModelServerConfig', (_message.Message,), {
  'DESCRIPTOR' : _MODELSERVERCONFIG,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:serving_config.ModelServerConfig)
  })
_sym_db.RegisterMessage(ModelServerConfig)


_MODELTYPE.values_by_name["MODEL_TYPE_UNSPECIFIED"]._options = None
_MODELTYPE.values_by_name["TENSORFLOW"]._options = None
_MODELTYPE.values_by_name["OTHER"]._options = None
_FILESYSTEMSTORAGEPATHSOURCECONFIG.fields_by_name['servable_name']._options = None
_FILESYSTEMSTORAGEPATHSOURCECONFIG.fields_by_name['base_path']._options = None
_FILESYSTEMSTORAGEPATHSOURCECONFIG.fields_by_name['fail_if_zero_versions_at_startup']._options = None
_MODELCONFIG_VERSIONLABELSENTRY._options = None
_MODELCONFIG.fields_by_name['model_type']._options = None
# @@protoc_insertion_point(module_scope)


class TFSConfig:

    @classmethod
    def from_model_configs(cls, model_configs: List[TFSModelConfig]) -> ModelServerConfig:
        serv_config = ModelServerConfig()
        model_config_list = serv_config.model_config_list
        for config in model_configs:
            model = model_config_list.config.add()
            model.name = config.name
            model.base_path = config.base_path
            model.model_platform = config.model_platform
            if config.model_versions:
                labels = model.version_labels
                policy = model.model_version_policy
                versions = policy.specific.versions
                for model_version in config.model_versions:
                    versions.append(model_version.step)
                    labels[model_version.label] = model_version.step
        return serv_config

    @classmethod
    def from_config_file(cls, filepath: str, as_obj: bool = True, **kwargs) -> Union[ModelServerConfig, List[TFSModelConfig]]:
        filepath = get_pathlike(filepath)
        assert filepath.exists()    
        serv_config = ModelServerConfig()
        serv_config = Merge(filepath.read_bytes(), serv_config)
        if not as_obj:
            return serv_config
    
        model_configs = []
        for model in serv_config.model_config_list.config:
            versions = [TFSModelVersion(step=v, label=k) for k,v in model.version_labels.items()]
            model_config = TFSModelConfig(name=model.name, base_path=model.base_path, model_platform=model.model_platform, model_versions=versions)
            model_configs.append(model_config)
        return model_configs
    
    @classmethod
    def from_protobuf(cls, filepath: str, as_obj: bool = True, **kwargs) -> Union[ModelServerConfig, List[TFSModelConfig]]:
        filepath = get_pathlike(filepath)
        assert filepath.exists()    
        serv_config = ModelServerConfig()
        serv_config.FromString(filepath.read_bytes())
        if not as_obj:
            return serv_config
        model_configs = []
        for model in serv_config.model_config_list.config:
            versions = [TFSModelVersion(step=v, label=k) for k,v in model.version_labels.items()]
            model_config = TFSModelConfig(name=model.name, base_path=model.base_path, model_platform=model.model_platform, model_versions=versions)
            model_configs.append(model_config)
        return model_configs


    @classmethod
    def from_dict(cls, configs: List[Dict[str, Any]], exclude=[], include=[], only_latest=False, return_tfconfig=False) -> Union[ModelServerConfig, List[TFSModelConfig]]:
        if exclude: configs = [c for c in configs if c.get('name', '') not in exclude]
        if include: configs = [c for c in configs if c.get('name', '') in include]
        model_configs = [ModelConfig.from_dict(config) for config in configs]
        if only_latest:
            for config in model_configs:
                config.set_only_latest()
        if not return_tfconfig:
            return model_configs
        return TFSConfig.from_model_configs(model_configs)
    
    @classmethod
    def to_config_file(cls, save_path: str, configs: List[Dict[str, Any]] = None, model_configs: Union[ModelServerConfig, List[TFSModelConfig]] = None, as_protobuf: bool = False, **kwargs):
        assert configs or model_configs, 'Configs or ModelConfigs must be provided'
        filepath = get_pathlike(save_path)
        srv_config = model_configs
        if configs:
            srv_config = TFSConfig.from_dict(configs, **kwargs)
        elif not isinstance(model_configs, ModelServerConfig):
            srv_config = TFSConfig.from_model_configs(model_configs, **kwargs)
        
        filepath.write_bytes(srv_config.SerializeToString()) if as_protobuf else filepath.write_text(str(srv_config))
        return {'config': srv_config, 'file': filepath}















from __future__ import annotations

"""
Extra Types for ORM
"""

from pathlib import Path
from sqlalchemy.sql import operators
from sqlalchemy.types import TypeDecorator, Text, String, DateTime
from sqlalchemy_json import NestedMutableJson, MutableDict, mutable_json_type
from sqlalchemy.dialects.postgresql import JSONB, ARRAY
from lazyops.utils.serialization import Json
from typing import Any, Optional, Union, Type

try:
    from fileio import File, FileLike
except ImportError:
    File = Path
    FileLike = Path


class FileField(TypeDecorator):

    impl = Text
    cache_ok = True

    def coerce_compared_value(self, op, value):
        """
        Coerce the value to be compared.
        """
        return Text() if op in (operators.like_op, operators.not_like_op) else self

    def process_bind_param(self, value: Optional[Union[str, FileLike]], dialect):
        """
        Return the value to be stored in the database.
        """
        if value is None: return None
        if isinstance(value, str): return value
        return value.as_posix() \
            if hasattr(value, 'as_posix') \
                else str(value)

    def process_result_value(self, value: Optional[Union[str, Any]], dialect):
        """
        Process the value retrieved from the database.
        """
        return File(value) if value is not None else None


class JsonString(TypeDecorator):
    """Enables JSON storage by encoding and decoding on the fly."""

    impl = String
    cache_ok = True

    def process_bind_param(self, value, dialect):
        """
        Bind the value to be stored in the database.
        """
        if value is not None: value = Json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        """
        Process the value retrieved from the database.
        """
        if value is not None: value = Json.loads(value)
        return value
    
    def coerce_compared_value(self, op, value):
        """
        Coerce the value to be compared.
        """
        return Text() \
            if op in (operators.like_op, operators.not_like_op) \
                else self


JsonStringField: Type[MutableDict] = mutable_json_type(dbtype = JsonString, nested=True)
JsonStringNestedField: Type[NestedMutableJson] = mutable_json_type(dbtype = JsonString, nested = True)


JSONBField: Type[MutableDict] = mutable_json_type(dbtype = JSONB)
JSONBNestedField: Type[NestedMutableJson] = mutable_json_type(dbtype = JSONB, nested = True)

DateTimeUTC: DateTime = DateTime(timezone = True)

StringList: ARRAY = ARRAY(Text, dimensions = 1)
DateTimeUTCList: ARRAY = ARRAY(DateTimeUTC, dimensions = 1)




from .base import BaseCompression
from ._gzip import GzipCompression
from ._lz4 import Lz4Compression
from ._zlib import ZlibCompression
from ._zstd import ZstdCompression
from typing import Any, Dict, Optional, Union, Type

CompressionT = Union[GzipCompression, Lz4Compression, ZlibCompression, ZstdCompression, BaseCompression]

def get_compression(
    compression_type: Optional[str] = None,
    compression_level: Optional[int] = None,
    **kwargs
) -> CompressionT:
    """
    Returns a Compression
    """
    compression_type = compression_type or "lz4"
    if compression_type == "gzip":
        return GzipCompression(compression_level = compression_level, **kwargs)
    if compression_type == "lz4":
        return Lz4Compression(compression_level = compression_level, **kwargs)
    if compression_type == "zlib":
        return ZlibCompression(compression_level = compression_level, **kwargs)
    if compression_type == "zstd":
        return ZstdCompression(compression_level = compression_level, **kwargs)
    raise ValueError(f"Invalid Compression Type: {compression_type}")


from __future__ import annotations

"""
Base ORM Objects
"""

from lazyops.imports._sqlalchemy import require_sql
require_sql(required=True, require_asyncpg=True, require_psycopg2=True)


import datetime
from uuid import UUID
from pydantic import BaseModel
from pydantic.alias_generators import to_snake
from dataclasses import dataclass
# from pydantic.dataclasses import dataclass

from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import as_declarative, declared_attr
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.orm import Mapped, InstrumentedAttribute
from sqlalchemy.orm import mapped_column
from sqlalchemy import func as sql_func
from sqlalchemy import Text
from sqlalchemy.orm import defer
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.inspection import inspect
from sqlalchemy.sql.expression import text, select, Select, ColumnElement, and_, update, Update, delete, or_
from sqlalchemy.dialects.postgresql import Insert, insert
from sqlalchemy.orm import selectinload
from sqlalchemy.orm import registry
from lazyops.utils.lazy import lazy_import
from lazyops.utils.logs import logger

from typing import Optional, Type, TypeVar, Union, Set, Any, Tuple, Literal, List, Dict, cast, Generic, Generator, Callable, TYPE_CHECKING

from . import errors


mapper_registry = registry()


class BasePydantic(MappedAsDataclass, DeclarativeBase, dataclass_callable = dataclass):
    
    if TYPE_CHECKING:
        id: Mapped[Union[str, int, UUID]]

    def __init__(self, **kw):
        mapper = inspect(self).mapper
        for key in mapper.relationships:
            if key in kw:
                kw[key] = mapper.relationships[key].entity.class_(**kw[key])
        super().__init__(**kw)
        self.orm_class_primary_keys = [c.name for c in mapper.primary_key]
        self.orm_class_primary_key = self.orm_class_primary_keys[0]

    @property
    def orm_class_name(self) -> str:
        """
        Returns the ORM class name
        """
        return self.__class__.__name__
    
    @property
    def orm_parent_class_name(self) -> str:
        """
        Returns the ORM parent class name
        """
        return self.__class__.__qualname__.split('.')[3]

    def get_non_relationship_fields(
        self,
        include: Optional[Set[str]] = None,
        exclude: Optional[Set[str]] = None,
        **kwargs
    ) -> List[str]:
        """
        Returns the non relationship fields
        """
        include = include or set()
        exclude = exclude or set()
        return [
            k for k in self.__dict__ if (
                k not in self.__mapper__.relationships or \
                k in include or \
                k not in exclude
            )
        ]
    
    def get_relationship_fields(
        self,
        include: Optional[Set[str]] = None,
        exclude: Optional[Set[str]] = None,
        **kwargs
    ) -> List[str]:
        """
        Returns the relationship fields
        """
        include = include or set()
        exclude = exclude or set()
        return [
            k for k in self.__mapper__.relationships if (
                k not in include or \
                k in exclude
            )
        ]

    def get_exportable_kwargs(
        self,
        include: Any = None,
        exclude: Any = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Returns the exportable kwargs
        """
        data = {
            k: v for k, v in self.__dict__.items() if k in self.get_non_relationship_fields(
                include = include,
                exclude = exclude,
                **kwargs,
            )
        }
        if exclude_none:
            data = {k: v for k, v in data.items() if v is not None}
        if exclude_unset or exclude_defaults:
            data = {k: v for k, v in data.items() if v != self.__mapper__.columns[k].default.arg}
        return data

SortOrder = Literal['ASC', 'DESC', 'asc', 'desc']
IDType = Union[str, int, UUID]
ModelTypeORM = TypeVar("ModelTypeORM", bound = BasePydantic)
SourceSchemaType = TypeVar("SourceSchemaType", bound = BaseModel)
CreateSchemaType = TypeVar("CreateSchemaType", bound = BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound = BaseModel)
SourceSchemaORM = TypeVar("SourceSchemaORM", bound = BaseModel)
ObjectResult = TypeVar("ObjectResult", BaseModel, BasePydantic)

async def has_this_obj(
    db: AsyncSession, 
    obj: ModelTypeORM, 
    attr: str, 
    obj_id: Union[int, str],
    primary_key: Optional[str] = None
) -> bool:
    """
    Returns whether or not the object has this object
    """
    await db.refresh(obj, [attr])
    primary_key = primary_key or obj.orm_class_primary_key
    return any(getattr(item, primary_key) == obj_id for item in getattr(obj, attr))


async def append_obj(
    db: AsyncSession,
    *,
    parent_obj: ModelTypeORM,
    child_obj: ModelTypeORM,
    attr: str
) -> Tuple[ModelTypeORM, ModelTypeORM]:
    """
    Appends the object to the parent object
    """
    await db.refresh(parent_obj, [attr])
    items: List[ModelTypeORM] = getattr(parent_obj, attr)
    items.append(child_obj)
    setattr(parent_obj, attr, items)

    await db.commit()
    await db.refresh(parent_obj)
    await db.refresh(child_obj)
    return parent_obj, child_obj


async def remove_obj(
    db: AsyncSession,
    *,
    parent_obj: ModelTypeORM,
    child_obj: ModelTypeORM,
    attr: str
) -> Tuple[ModelTypeORM, ModelTypeORM]:
    """
    Removes the object from the parent object
    """
    await db.refresh(parent_obj, [attr])
    items: List[ModelTypeORM] = getattr(parent_obj, attr)
    items.remove(child_obj)
    setattr(parent_obj, attr, items)

    await db.commit()
    await db.refresh(parent_obj)
    await db.refresh(child_obj)
    return parent_obj, child_obj


def prepare_update_data(
    model: Type[ModelTypeORM], 
    update_data: Union[SourceSchemaType, Dict[str, Any]]
) -> Dict[str, Any]:
    """
    Prepares the update data
    """
    if isinstance(update_data, BaseModel):
        update_data = update_data.model_dump(exclude_unset=True)
    update_data = {k: v for k, v in update_data.items() if hasattr(model, k)}
    if hasattr(model, 'updated_at'):
        update_data['updated_at'] = datetime.datetime.now(datetime.timezone.utc)
    return update_data



class CRUDBase(Generic[ModelTypeORM, SourceSchemaType]):

    def __init__(
        self, 
        model: Union[Type[ModelTypeORM], str], 
        defer_attrs: List[str] = None,
        auto_commit: Optional[bool] = True,
        primary_key: Optional[str] = None,
        model_dump_kwargs: Optional[Dict[str, Any]] = None,
        source_model: Optional[Type[SourceSchemaType]] = None,
        **kwargs
    ):
        """
        CRUD object with default methods to Create, Read, Update, Delete (CRUD).

        **Parameters**

        * `model`: A SQLAlchemy model class
        * `schema`: A Pydantic model (schema) class
        * `defer_attrs`: A list of attributes to defer loading
        """
        if isinstance(model, str): model = lazy_import(model)
        self.model: Type[ModelTypeORM] = model
        self.defer_attrs = defer_attrs or []
        self.auto_commit = auto_commit
        self._kwargs = kwargs
        if not primary_key:
            primary_key = inspect(self.model).primary_key[0].name
        self.primary_key = primary_key
        self.object_primary_key_field: InstrumentedAttribute[Union[str, int, UUID]] = getattr(self.model, self.primary_key)
        self.model_dump_kwargs = model_dump_kwargs or {}
        if source_model:
            if isinstance(source_model, str): source_model = lazy_import(source_model)
        self.source_model: Optional[Type[SourceSchemaType]] = source_model

    @property
    def table_name(self) -> str:
        """
        Returns the table name
        """
        return self.model.__tablename__

    @property
    def object_class_name(self) -> str:
        """
        Returns the object class name
        """
        return self.model.__name__
    

    """
    Base Methods
    """

    async def all(
        self, 
        db: AsyncSession, 
        stmt: Select,
        **kwargs
    ) -> List[ObjectResult]:
        """
        Returns all objects of the query
        """
        res = (await db.scalars(stmt)).all()
        res = cast(List[ModelTypeORM], res)
        return res
    
    async def first(
        self, 
        db: AsyncSession, 
        stmt: Union[Select, Insert, Update], 
        raise_for_not_found: bool = False,
        **kwargs
    ) -> Optional[ObjectResult]:
        """
        Returns the first object of the query
        """
        res = (await db.scalars(stmt)).first()
        res = cast(self.model, res)
        self._raise_for_not_found(res, raise_for_not_found)
        return res

    """
    Get / Read Methods
    """

    async def get(
        self,
        db: AsyncSession,
        id: IDType,
        raise_for_not_found: bool = False,
        defer_attrs: List[str] = None,
        **kwargs
    ) -> Optional[ObjectResult]:
        """
        Returns the object with the given id
        """
        stmt = select(self.model).where(self.object_primary_key_field == id)
        stmt = self._defer_attrs(stmt, defer_attrs)
        return await self.first(db, stmt, raise_for_not_found)


    async def get_first_filtered(
        self, 
        db: AsyncSession, 
        raise_for_not_found: bool = False, 
        **filters
    ) -> Optional[ObjectResult]:
        """
        Gets the first object with the given filters
        """
        and_stmt = self._construct_and_statement(filters)
        stmt = select(self.model).where(and_stmt)
        stmt = self._defer_attrs(stmt, None)
        return await self.first(db, stmt, raise_for_not_found)

    async def get_multi(
        self, 
        db: AsyncSession, 
        *, 
        skip: int = 0, 
        limit: int = 100,
        sort_field: Optional[str] = None,
        sort_order: Optional[SortOrder] = None,
        **filters
    ) -> List[ObjectResult]:
        """
        Gets multiple objects with pagination
        """
        and_stmt = self._construct_and_statement(filters)
        stmt = select(self.model)
        if and_stmt: stmt = stmt.where(and_stmt)
        if skip is not None: stmt = stmt.offset(skip)
        if limit is not None: stmt = stmt.limit(limit)
        if sort_field: stmt = self._construct_sort_statement(stmt, sort_field, sort_order)
        stmt = self._defer_attrs(stmt, None)
        return await self.all(db, stmt)

    
    async def get_multi_ids(
        self,
        db: AsyncSession,
        *,
        ids: List[IDType],
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort_field: Optional[str] = None,
        sort_order: Optional[SortOrder] = None,
        load_children: List[str] = None,
        raise_for_not_found: bool = False,
        **kwargs
    ) -> List[ObjectResult]:
        """
        Gets multiple objects with pagination and filters by ids
        """
        ids_string = self._create_id_string(ids)
        stmt: Select = select(self.model).where(text(f'{self.primary_key} in ({ids_string})'))
        if skip is not None: stmt = stmt.offset(skip)
        if limit is not None: stmt = stmt.limit(limit)
        if sort_field: stmt = self._construct_sort_statement(stmt, sort_field, sort_order)
        if load_children: stmt = self._add_related_obj_loading(stmt, load_children)
        objs = await self.all(db, stmt)
        if raise_for_not_found and len(objs) < len(ids):
            raise errors.MissingItemsException(self.object_class_name)
        return objs
    

    async def get_random(
        self, 
        db: AsyncSession, 
        *, 
        limit: int = 100, 
        sort_field: Optional[str] = None,
        sort_order: Optional[SortOrder] = None,
        **filters
    ) -> List[ObjectResult]:
        """
        Gets multiple objects with pagination and filters
        """
        and_stmt = self._construct_and_statement(filters)
        stmt = select(self.model)
        if and_stmt: stmt = stmt.where(and_stmt)
        stmt = stmt.order_by(sql_func.random()).limit(limit)
        if sort_field: stmt = self._construct_sort_statement(stmt, sort_field, sort_order)
        stmt = self._defer_attrs(stmt, None)
        return await self.all(db, stmt)
    

    async def get_count(
        self, 
        db: AsyncSession, 
        **filters
    ) -> int:
        """
        Gets the count of objects with filters
        """
        and_stmt = self._construct_and_statement(filters)
        stmt = select(sql_func.count(self.object_primary_key_field)) # .where(and_stmt)
        if and_stmt: stmt = stmt.where(and_stmt)
        return await db.scalar(stmt)
    
    """
    Create Methods
    """

    async def create(
        self, 
        db: AsyncSession, 
        *, 
        obj_in: SourceSchemaType,
        model_dump_kwargs: Optional[Dict[str, Any]] = None,
        no_commit: Optional[bool] = None,
        **kwargs
    ) -> ModelTypeORM:
        """
        Creates an object
        """
        db_obj = self._convert_source_to_orm_model(obj_in, model_dump_kwargs = model_dump_kwargs)
        if hasattr(db_obj, 'created_at'):
            db_obj.created_at = datetime.datetime.now(datetime.timezone.utc)
        db.add(db_obj)
        do_commit = no_commit is False if no_commit is not None else self.auto_commit
        if do_commit:
            await db.commit()
            await db.refresh(db_obj)
        return db_obj
    

    async def create_if_not_exists(
        self, 
        db: AsyncSession, 
        *, 
        obj_in: SourceSchemaType, 
        index_elements: Optional[List[str]] = None,
        model_dump_kwargs: Optional[Dict[str, Any]] = None,
        exclude_attrs: Optional[List[str]] = None,
        no_commit: Optional[bool] = None,
        **kwargs
    ) -> None:
        """
        Creates an object if it does not exist
        """
        index_elements = index_elements or [self.primary_key]
        values = self._convert_source_to_dict(
            obj_in, 
            model_dump_kwargs = model_dump_kwargs,
            exclude = exclude_attrs,
            **kwargs
        )
        stmt = insert(self.model).values(
            **values
        ).on_conflict_do_nothing(
            index_elements = index_elements
        )
        await db.execute(stmt)
        do_commit = no_commit is False if no_commit is not None else self.auto_commit
        if do_commit:
            await db.commit()

    """
    Upsert Methods
    """

    async def upsert(
        self, 
        db: AsyncSession, 
        *, 
        obj_in: SourceSchemaType, 
        model_dump_kwargs: Optional[Dict[str, Any]] = None,
        no_commit: Optional[bool] = None,
        **kwargs
    ) -> None:
        """
        Upserts an object
        """
        values = self._convert_source_to_dict(
            obj_in, 
            model_dump_kwargs = model_dump_kwargs,
            **kwargs
        )
        stmt = insert(self.model).values(
            **values
        ).on_conflict_do_update(
            index_elements=[self.primary_key],
            set_=values
        )
        await db.execute(stmt)
        do_commit = no_commit is False if no_commit is not None else self.auto_commit
        if do_commit:
            await db.commit()


    async def upsert_multi(
        self,
        db: AsyncSession,
        *,
        objs_in: List['SourceSchemaType'],
        model_dump_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        """
        Upserts multiple objects
        """
        for obj_in in objs_in:
            await self.upsert(db, obj_in = obj_in, model_dump_kwargs = model_dump_kwargs, no_commit = True)
        await db.commit()


    async def upsert_bulk(
        self,
        db: AsyncSession,
        *,
        objs_in: List['SourceSchemaType'],
        model_dump_kwargs: Optional[Dict[str, Any]] = None,
        batch_size: Optional[int] = 200,
        **kwargs
    ) -> int:
        """
        Upserts multiple objects
        """
        for i in range(0, len(objs_in), batch_size):
            await self.upsert_multi(db, objs_in = objs_in[i:i+batch_size], model_dump_kwargs = model_dump_kwargs, **kwargs)

    async def upsert_multi_with_idx(
        self,
        db: AsyncSession,
        *,
        objs_in: List['SourceSchemaType'],
        start_idx: int,
        model_dump_kwargs: Optional[Dict[str, Any]] = None,
        raise_errors: Optional[bool] = None,
        **kwargs
    ) -> Tuple[List[int], List[int]]:
        """
        Upserts multiple objects with an index
        """
        completed_idx: List[int] = []
        failed_idx: List[int] = []
        for n, obj_in in enumerate(objs_in):
            try:
                await self.upsert(db, obj_in = obj_in, no_commit = True, model_dump_kwargs = model_dump_kwargs)
                completed_idx.append(start_idx + n)
            except Exception as e:
                logger.info(f'|r|Attempting to rollback on error|e|: {e}', colored = True, prefix = f'|r|[{getattr(obj_in, self.primary_key)} - {start_idx + n}] Error in upserting data|e|')
                await db.rollback()
                try:
                    await self.upsert(db, obj_in = obj_in, no_commit = True)
                    completed_idx.append(start_idx + n)
                except Exception as e:
                    logger.info(obj_in, colored = True, prefix = f'|r|[{getattr(obj_in, self.primary_key)} - {start_idx + n}] Error in upserting data|e|')
                    logger.trace('Error in upserting data', e)
                    if raise_errors: raise e
                    failed_idx.append(start_idx + n)
        await db.commit()
        return (completed_idx, failed_idx)


    async def upsert_bulk_with_index(
        self,
        db: AsyncSession,
        *,
        objs_in: List['SourceSchemaType'],
        batch_size: Optional[int] = 200,
        model_dump_kwargs: Optional[Dict[str, Any]] = None,
        raise_errors: Optional[bool] = None,
        **kwargs
    ) -> Tuple[List[int], List[int]]:
        """
        Upserts multiple objects
        """
        completed_idx: List[int] = []
        failed_idx: List[int] = []
        for i in range(0, len(objs_in), batch_size):
            idxs = await self.upsert_multi_with_idx(
                db, 
                objs_in = objs_in[i:i+batch_size], 
                start_idx = i, 
                raise_errors = raise_errors,
                model_dump_kwargs = model_dump_kwargs,
                **kwargs
            )
            completed, failed = idxs
            if completed: completed_idx += completed
            if failed: failed_idx += failed
        return (completed_idx, failed_idx)

    """
    Update Methods
    """
    
    async def update(
        self,
        db: AsyncSession,
        *,
        db_obj: ModelTypeORM,
        obj_in: Union[SourceSchemaType, Dict[str, Any]],
        **kwargs
    ) -> ModelTypeORM:
        """
        Updates an object
        """
        update_data = prepare_update_data(self.model, obj_in)
        for field, value in update_data.items():
            setattr(db_obj, field, value)
        db.add(db_obj)
        if self.auto_commit: 
            await db.commit()
            await db.refresh(db_obj)
        return db_obj

    async def update_by_filter(
        self, 
        db: AsyncSession, 
        obj_in: Union[SourceSchemaType, Dict], 
        **filters
    ) -> Dict:
        """
        Updates an object by filter
        """
        update_data = prepare_update_data(self.model, obj_in)
        and_stmt = self._construct_and_statement(filters)
        returning_stmt = [getattr(self.model, attr) for attr in self._usable_attributes()]
        stmt = update(self.model).where(and_stmt).values(update_data).returning(*returning_stmt)
        res = (await db.execute(stmt)).first()
        self._raise_for_not_found(res, raise_for_not_found=True)
        if self.auto_commit: await db.commit()
        return dict(zip(self._usable_attributes(), res))
    
    """
    Deletion Methods
    """

    async def delete_obj(self, db: AsyncSession, *, db_obj: ModelTypeORM) -> ModelTypeORM:
        """
        Removes an object
        """
        await db.delete(db_obj)
        await db.commit()
        return db_obj

    async def delete(
        self, 
        db: AsyncSession, 
        *, 
        id: IDType
    ) -> ModelTypeORM:
        """
        Removes an object
        """
        obj = await self.get(db, id=id, raise_for_not_found=True)
        return await self.delete_obj(db, db_obj=obj)

    async def delete_all_filtered(self, db: AsyncSession, **filters) -> None:
        """
        Deletes all objects with the given filters
        """
        and_stmt = self._construct_and_statement(filters)
        stmt = delete(self.model).where(and_stmt)
        await db.execute(stmt)
        await db.commit()


    """
    Utility Methods
    """

    async def get_from_results(
        self,
        db: AsyncSession,
        *,
        results: List[Union[Tuple[str, Any], IDType]],
        cast_to_source: Optional[bool] = None,
        source_model: Optional[Type[SourceSchemaType]] = None,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort_field: Optional[str] = None,
        sort_order: Optional[SortOrder] = None,
        **kwargs
    )  -> List[ObjectResult]:
        """
        Fetches the objects from the results and optionally casts them to a source model
        """
        id_list = []
        for result in results:
            if isinstance(result, tuple):
                id_list.append(result[0])
            else:
                id_list.append(result)
        objs = await self.get_multi_ids(db, ids = id_list, skip = skip, limit = limit, sort_field = sort_field, sort_order = sort_order, **kwargs)
        if cast_to_source:
            source_model = source_model or self.source_model
            objs = [self._convert_orm_model_to_source(obj, source_model) for obj in objs]
        return objs

    
    def _create_id_string(self, ids: List[IDType]) -> str:
        """
        Creates an id string
        """
        _ids = []
        for id in ids:
            if isinstance(id, str):
                _ids.append(f"'{id}'")
            elif isinstance(id, UUID):
                _ids.append(f"'{str(id)}'")
            else:
                _ids.append(str(id))
        return ', '.join(_ids)

    def _add_related_obj_loading(self, stmt: Select, load_children: List[str] = None) -> Select:
        """
        Constructs a select statement with the given related objects
        """
        if load_children:
            options = iter(getattr(self.model, attr) for attr in load_children)
            stmt = stmt.options(selectinload(*options))
        return stmt

    def _raise_for_not_found(self, obj: Optional[ModelTypeORM], raise_for_not_found: bool) -> None:
        """
        Raises an error if the object is not found
        """
        if not obj and raise_for_not_found:
            raise errors.NoItemFoundException(self.object_class_name)


    def _defer_attrs(self, stmt: Union[Select, Update, Insert], defer_attrs: Optional[List[str]]):
        """
        Returns the statement with the deferred attributes
        """
        defer_attrs = defer_attrs or []
        if self.defer_attrs: defer_attrs += self.defer_attrs
        if defer_attrs:
            stmt = stmt.options(*[defer(getattr(self.model, attr), raiseload=True) for attr in defer_attrs])
        return stmt

    def _construct_and_statement(self, filters: dict) -> ColumnElement:
        """
        Constructs an AND statement from the given filters
        """
        if not filters: return None
        _filters = ()
        for attr, value in filters.items():
            _filters += (getattr(self.model, attr) == value,)
        return and_(*_filters)

    def _construct_or_statement(self, filters: dict) -> ColumnElement:
        """
        Constructs an OR statement from the given filters
        """
        if not filters: return None
        _filters = ()
        for attr, value in filters.items():
            _filters += (getattr(self.model, attr) == value,)
        return or_(*_filters)

    def _construct_sort_statement(
        self, 
        stmt: Select, 
        sort_field: str, 
        sort_order: Optional[SortOrder] = None
    ) -> ColumnElement:
        """
        Constructs a sort statement from the given sort field and sort order
        """
        if not sort_order: sort_order = 'ASC'
        sort_order = sort_order.upper()
        return stmt.order_by(getattr(self.model, sort_field).asc() if sort_order == 'ASC' else getattr(self.model, sort_field).desc())
    
    def _convert_source_to_dict(
        self,
        source: SourceSchemaType,
        model_dump_kwargs: Optional[Dict[str, Any]] = None,
        use_encoder: Optional[bool] = True,
        **kwargs
    ) -> Dict[str, Any]:
        """
        [model_dump | jsonable_encoder] Converts the source to a dict
        """
        dump_kwargs = self.model_dump_kwargs.copy()
        if model_dump_kwargs: dump_kwargs.update(model_dump_kwargs)
        if kwargs:
            for k, v in kwargs.items():
                if v is None: continue
                if k not in dump_kwargs:
                    dump_kwargs[k] = v
                    continue
                if isinstance(dump_kwargs[k], set):
                    dump_kwargs[k].add(v)
                elif isinstance(dump_kwargs[k], list):
                    dump_kwargs[k].append(v)
                elif isinstance(dump_kwargs[k], dict):
                    dump_kwargs[k].update(v)
                else:
                    dump_kwargs[k] = v
        if use_encoder:
            return jsonable_encoder(source, **dump_kwargs)
        return source.model_dump(**dump_kwargs)
    
    def _convert_source_to_orm_model(
        self,
        source: SourceSchemaType,
        model_dump_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ModelTypeORM:
        """
        Converts the source to the ORM model
        """
        data = self._convert_source_to_dict(source, model_dump_kwargs = model_dump_kwargs, **kwargs)
        try:
            return self.model(**data)
        except Exception as e:
            logger.trace(f'Error converting {type(source)} -> {self.object_class_name}', e)
            raise e

    def _convert_orm_model_to_source(
        self,
        obj: ObjectResult,
        source_model: Type[SourceSchemaType],
        **kwargs
    ) -> SourceSchemaType:
        """
        Converts the ORM model to the source model
        """
        values: Dict[str, Union[Dict[str, Any], Any]] = obj.model_dump(**kwargs)
        field_names, field_aliases = [], {}
        for name, field in source_model.model_fields.items():
            field_names.append(name)
            if field.alias: field_aliases[field.alias] = name
        if not isinstance(values, dict): 
            values = values.__dict__
            _ = values.pop('_sa_instance_state', None)
        
        # Handle Incoming Metadata
        if 'metadata_' in values and 'metadata' not in values:
            values['metadata'] = values.pop('metadata_')
        
        value_keys = set(values.keys())
        metadata = {
            key: values.pop(key)
            for key in value_keys
            if key not in field_names and key not in field_aliases
        }
        _ = metadata.pop('_sa_instance_state', None)
        if not values.get('metadata'):
            values['metadata'] = metadata
        else:
            metadata = {k: v for k, v in metadata.items() if v is not None}
            values['metadata'].update(metadata)
        for key, value in values['metadata'].items():
            if value is None: continue
            if key in field_names:
                values[key] = value
                continue
            if key in field_aliases:
                values[field_aliases[key]] = value
                continue
        return source_model.model_validate(values)

    def _usable_attributes(self) -> Generator[str, None, None]:
        """
        Returns the usable attributes
        """
        for attr in inspect(self.model).c.keys():
            if not self.defer_attrs or attr not in self.defer_attrs:
                yield attr



from __future__ import annotations

from pathlib import Path
from lazyops.types import BaseModel, lazyproperty, validator, Field
from lazyops.configs.base import DefaultSettings, BaseSettings
from typing import List, Dict, Union, Any, Optional

class KOpsSettings(DefaultSettings):
    kubeconfig: Optional[str] = None
    kubeconfig_context: Optional[str] = None
    kops_ctx_dir: Optional[str] = None
    kops_ctx: Optional[str] = None

    kops_watch_interval: int = 60
    kops_finalizer: str = 'kops.gex.ai/finalizer'
    kops_prefix: str = 'kops.gex.ai'
    kops_persistent_key: str = 'last-handled-configuration'
    kops_generate_event_name: str = 'kops-event-'

    kops_max_message_length: int = 1024
    kops_cut_message_infix: str = '...'

    kopf_enable_event_logging: bool = False
    kopf_event_logging_level: str = 'INFO'

    kopf_name: str = 'KOps'
    build_id: Optional[str] = None


    class Config:
        arbitrary_types_allowed = True
        case_sensitive = False
        env_prefix = ''

    @lazyproperty
    def kubeconfig_path(self):
        if not self.kubeconfig: return None
        path = Path(self.kubeconfig)
        return path if path.exists() else None

    @lazyproperty
    def kops_ctx_path(self):
        return Path(self.kops_ctx_dir) if self.kops_ctx_dir else None
    
    @lazyproperty
    def kops_kconfig_path(self):
        if self.kops_ctx and self.kops_ctx_path:
            for ext in {'.yaml', '.yml'}:
                if ext in self.kops_ctx: ext = ''
                path = self.kops_ctx_path.joinpath(f'{self.kops_ctx}{ext}')
                if path.exists():
                    return path
        return None
        
    @lazyproperty
    def kconfig_path(self) -> Path:
        return self.kops_kconfig_path or self.kubeconfig_path 
    
    def get_kconfig_path(self, ctx: str = None) -> Path:
        if ctx:
            for ext in {'.yaml', '.yml'}:
                if ext in ctx: ext = ''
                path = self.kops_ctx_path.joinpath(f'{ctx}{ext}')
                if path.exists():
                    return path

        return self.kconfig_path


"""
Module Import Handler
"""
import typing
import functools
import pathlib
import importlib.util
from lazyops.utils.lazy import lazy_import

from typing import Optional, Dict, List, Union, Any, Callable, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from lazyops.types import BaseModel


@functools.lru_cache()
def get_module_path(
    module_name: str, 
    **kwargs
) -> pathlib.Path:
    """
    Get the path to the module.

    args:
        module_name: name of the module to import from (e.g. 'lazyops')
    
    Use it like this:

    >>> get_module_path('lazyops')
    """
    module_spec = importlib.util.find_spec(module_name)
    if not module_spec:
        raise ValueError(f"Module {module_name} not found")
    
    for path in module_spec.submodule_search_locations:
        module_path = pathlib.Path(path)
        if module_path.exists(): return module_path
    
    raise ValueError(f"Module {module_name} cant be found in the path")


@functools.lru_cache()
def get_module_assets_path(
    module_name: str, 
    assets_dir: typing.Optional[str] = 'assets',
    **kwargs
) -> pathlib.Path:
    """
    Get the path to the module's assets.

    args:
        module_name: name of the module to import from (e.g. 'lazyops')
        assets_dir: name of the assets directory (default: 'assets')
    
    Use it like this:

    >>> get_module_assets_path('lazyops', assets_dir = 'assets')
    """
    module_spec = importlib.util.find_spec(module_name)
    if not module_spec:
        raise ValueError(f"Module {module_name} not found")
    
    for path in module_spec.submodule_search_locations:
        asset_path = pathlib.Path(path).joinpath(assets_dir)
        if asset_path.exists(): return asset_path
    
    raise ValueError(f"Module {module_name} does not have an assets directory")


_allowed_extensions = [
    '.yaml', '.yml', '.json', '.jsonl', '.jsonlines', '.txt',
    '.csv', '.tsv', '.xls', '.xlsx', '.xlsm', '.xlsb', '.odf', '.ods', '.odt',
    '.pdf', '.doc', '.docx', '.ppt', '.pptx', '.rtf', '.txt', '.md', '.html', '.htm',
    '.pkl', '.pickle',
    '.jpg', '.jpeg', '.png', '.gif', '.bmp', '.tiff', '.tif', '.svg', '.webp',
]

_binary_extensions = [
    '.pkl',
    '.pickle',
    '.pdf',
    '.doc',
    '.docx',
    '.jpg',
    '.jpeg',
    '.png',
    '.gif',
    '.bmp',
    '.tiff',
    '.tif',
    '.svg',
    '.webp',
]


@functools.lru_cache()
def search_for_assets(
    path: pathlib.Path,
    allowed_extensions: typing.Optional[typing.List[str]] = None,
    recursive: typing.Optional[bool] = False,
    **kwargs
) -> typing.List[pathlib.Path]:
    """
    Search for assets in a path.

    args:
        path: path to search
        allowed_extensions: list of allowed extensions (default: None)
        recursive: search recursively (default: False)
    
    Use it like this:

    >>> search_for_assets(pathlib.Path('lazyops/assets'), allowed_extensions = ['.yaml', '.yml'])
    """
    if allowed_extensions is None: allowed_extensions = _allowed_extensions
    result = []
    for file in path.iterdir():
        if file.is_dir() and recursive:
            result.extend(search_for_assets(file, allowed_extensions = allowed_extensions))
        elif file.suffix in allowed_extensions:
            result.append(file)
    return result

_file_loader_extensions = {
    'fileio.io.Yaml.loads': ['.yaml', '.yml'],
    'lazyops.utils.serialization.Json.loads': ['.json', '.jsonl', '.jsonlines'],
    'fileio.io.Csv': ['.csv'],
    'fileio.io.Tsv': ['.tsv'],
    'fileio.io.Dill': ['.pkl', '.pickle'],
}

def get_file_loader(
    path: pathlib.Path,
):
    """
    Get the file loader for the path.
    """
    return next(
        (
            lazy_import(loader)
            for loader, extensions in _file_loader_extensions.items()
            if path.suffix in extensions
        ),
        None,
    )


@functools.lru_cache()
def get_module_assets(
    module_name: str, 
    *path_parts,
    assets_dir: Optional[str] = 'assets',
    allowed_extensions: Optional[List[str]] = None,
    recursive: Optional[bool] = False,
    **kwargs
) -> Union[pathlib.Path, List[pathlib.Path]]:
    """
    Get the path to an assets from a module.

    args:
        module_name: name of the module to import from (e.g. 'configz')
        path_parts: path parts to the assets directory (default: [])
        assets_dir: name of the assets directory (default: 'assets')
        allowed_extensions: list of allowed extensions (default: None)
    
    Use it like this:

    >>> get_module_assets_path('lazyops', 'authz', 'file.json', assets_dir = 'assets')
    >>> get_module_assets_path('lazyops', 'authz') 
    """
    module_assets_path = get_module_assets_path(module_name, assets_dir = assets_dir)
    module_assets = module_assets_path.joinpath(*path_parts)
    # if its a file, return the file
    if module_assets.is_file(): return module_assets
    # if it's a directory, return a list of files
    return search_for_assets(module_assets, allowed_extensions = allowed_extensions, recursive = recursive)


@functools.lru_cache()
def load_file_content(
    path: pathlib.Path,
    model: Optional[Type['BaseModel']] = None,
    loader: Optional[Callable] = None,
    binary_load: Optional[bool] = None,
    **kwargs
) -> Any:
    """
    Load file content.

    args:
        path: path to the file
        model: model to parse the file with (default: None)
        loader: loader to use (default: None)
        binary_load: load the file as binary (default: None)
    """
    assert path.exists(), f"File {path} does not exist"
    binary_load = path.suffix in _binary_extensions if binary_load is None else binary_load
    if loader is None:
        loader = get_file_loader(path)
    data = path.read_bytes() if binary_load else path.read_text()
    if loader is None and model is None: return data
    if loader is not None:
        if path.suffix in {'jsonl', 'jsonlines'}:
            data = [loader(d) for d in data.splitlines()]
        else:
            data = loader(data)
    # load model
    if model is not None: 
        from lazyops.types.models import pyd_parse_obj
        data = pyd_parse_obj(model, data)
    return data


@functools.lru_cache()
def import_module_assets(
    module_name: str, 
    *path_parts,
    model: Optional[Type['BaseModel']] = None,
    load_file: Optional[bool] = False,
    loader: Optional[Callable] = None,
    assets_dir: Optional[str] = 'assets',
    allowed_extensions: Optional[List[str]] = None,
    recursive: Optional[bool] = False,
    as_dict: Optional[bool] = False,
    **kwargs
) -> Dict[str, Any]:
    """
    Import assets from a module.

    args:
        module_name: name of the module to import from (e.g. 'configz')
        path_parts: path parts to the assets directory (default: [])
        model: model to parse the assets with (default: None)
        load_file: load the file (default: False)
        loader: loader to use (default: None)
        assets_dir: name of the assets directory (default: 'assets')
        allowed_extensions: list of allowed extensions (default: None)
        as_dict: return a dict instead of a list (default: False)
    """

    module_assets = get_module_assets(
        module_name,
        *path_parts,
        assets_dir = assets_dir,
        allowed_extensions = allowed_extensions,
        recursive = recursive,
        **kwargs
    )
    if not isinstance(module_assets, list): module_assets = [module_assets]
    return {
        asset.name: load_file_content(asset, model=model, loader=loader, **kwargs)
        if load_file or model is not None
        else asset
        for asset in module_assets
    } if as_dict else [
        load_file_content(asset, model=model, loader=loader, **kwargs)
        if load_file or model is not None
        else asset
        for asset in module_assets
    ]





"""
Wrapper Functions
"""

def create_import_assets_wrapper(
    module_name: str,
    assets_dir: Optional[str] = 'assets',
    allowed_extensions: Optional[List[str]] = None,
    recursive: Optional[bool] = False,
    as_dict: Optional[bool] = False,
) -> Callable[..., Union[Dict[str, Any], List[Any]]]:
    """
    Create a wrapper around import_module_assets.

    args:
        module_name: name of the module to import from (e.g. 'lazyops')
        assets_dir: name of the assets directory (default: 'assets')
        allowed_extensions: list of allowed extensions (default: None)
        as_dict: return a dict instead of a list (default: False)

    """
    def import_assets_wrapper(
        *path_parts,
        model: Optional[Type['BaseModel']] = None,
        load_file: Optional[bool] = False,
        **kwargs
    ) -> Union[Dict[str, Any], List[Any]]:
        """
        Import assets from a module.

        args:
            path_parts: path parts to the assets directory (default: [])
            model: model to parse the assets with (default: None)
            load_file: load the file (default: False)
            **kwargs: additional arguments to pass to import_module_assets
        
        """
        return import_module_assets(
            module_name,
            *path_parts,
            assets_dir = assets_dir,
            model = model,
            load_file = load_file,
            allowed_extensions = allowed_extensions,
            recursive = recursive,
            as_dict = as_dict,
            **kwargs,
        )

    return import_assets_wrapper


def create_get_assets_wrapper(
    module_name: str,
    assets_dir: Optional[str] = 'assets',
    allowed_extensions: Optional[List[str]] = None,
    recursive: Optional[bool] = False,
    as_dict: Optional[bool] = False,
) -> Callable[..., Union[pathlib.Path, Any, List[pathlib.Path], List[Any], Dict[str, pathlib.Path], Dict[str, Any]]]:
    """
    Create a wrapper around `get_module_assets`.

    args:
        module_name: name of the module to import from (e.g. 'lazyops')
        assets_dir: name of the assets directory (default: 'assets')
        allowed_extensions: list of allowed extensions (default: None)
        as_dict: return a dict instead of a list (default: False)

    """
    def get_assets_wrapper(
        *path_parts,
        load_file: Optional[bool] = False,
        loader: Optional[Callable] = None,
        **kwargs
    ) -> Union[pathlib.Path, Any, List[pathlib.Path], List[Any], Dict[str, pathlib.Path], Dict[str, Any]]:
        """
        Import assets from a module.

        args:
            path_parts: path parts to the assets directory (default: [])
            load_file: load the file (default: False)
            loader: loader function to use (default: None)
            **kwargs: additional arguments to pass to `get_module_assets`
        
        """
        assets = get_module_assets(
            module_name,
            *path_parts,
            assets_dir = assets_dir,
            allowed_extensions = allowed_extensions,
            recursive = recursive,
            **kwargs,
        )
        if as_dict:
            if not isinstance(assets, list): assets = [assets]
            return {
                asset.name: load_file_content(asset, loader = loader, **kwargs) if load_file else asset
                for asset in assets
            }
        
        if isinstance(assets, list):
            return [load_file_content(asset, loader = loader, **kwargs) if load_file else asset for asset in assets]
        return load_file_content(assets, loader = loader, **kwargs) if load_file else assets

    return get_assets_wrapper

